import logging
import inspect
from typing import Dict, List, Optional, Any
from functools import wraps
import inspect
from opentelemetry.context import attach, get_current, detach
from opentelemetry.sdk.trace import Span
from opentelemetry.sdk.trace import Span
from opentelemetry.trace import get_tracer
from opentelemetry.trace.propagation import set_span_in_context, _SPAN_KEY
from contextlib import contextmanager, asynccontextmanager
from monocle_apptrace.instrumentation.common.span_handler import SpanHandler
from monocle_apptrace.instrumentation.common.wrapper import atask_wrapper, task_wrapper
from monocle_apptrace.instrumentation.common.utils import (
    set_scope, remove_scope, http_route_handler, http_async_route_handler
)
from monocle_apptrace.instrumentation.common.constants import MONOCLE_INSTRUMENTOR
from monocle_apptrace.instrumentation.common.instrumentor import get_tracer_provider

logger = logging.getLogger(__name__)




def start_trace(
    span_name: Optional[str] = None,
    attributes: Optional[Dict[str, Any]] = None,
    events: Optional[List[Dict[str, Any]]] = None
):
    """
    Starts a new trace. All the spans created after this call will be part of the same trace. 
    
    Args:
        span_name: Optional custom span name. If None, uses the default span name.
        attributes: Optional dictionary of custom attributes to set on the span.
        events: Optional list of events to add to the span. Each event should be a dict with 
                'name' and optionally 'attributes' keys.
    
    Returns:
        Token: A token representing the attached context for the span.
                      This token is to be used later to stop the current trace.
                      Returns None if tracing fails.
    
    Raises:
        Exception: The function catches all exceptions internally and logs a warning.
    """
    try:
        tracer = get_tracer(instrumenting_module_name= MONOCLE_INSTRUMENTOR, tracer_provider= get_tracer_provider())
        span_name = span_name or "custom_span"
        span = tracer.start_span(name=span_name)
        updated_span_context = set_span_in_context(span=span)
        
        # Set default monocle attributes
        SpanHandler.set_default_monocle_attributes(span)
        if SpanHandler.is_root_span(span):
            SpanHandler.set_workflow_properties(span)
        
        # Set custom attributes and events using common method
        _setup_span_attributes_and_events(span, attributes, events)
        
        token = attach(updated_span_context)
        return token
    except Exception as e:
        logger.warning(f"Failed to start trace: {e}")
        return None

def stop_trace(
    token,
    final_attributes: Optional[Dict[str, Any]] = None,
    final_events: Optional[List[Dict[str, Any]]] = None
) -> None:
    """
    Stop the active trace. All the spans created after this will not be part of the trace.
    
    Args:
        token: The token that was returned when the trace was started. Can be None in which case only the span is ended.
        final_attributes: Optional dictionary of final attributes to set on the span before ending.
        final_events: Optional list of final events to add to the span before ending.
    
    Returns:
        None
    """
    try:
        _parent_span_context = get_current()
        if _parent_span_context is not None:
            parent_span: Span = _parent_span_context.get(_SPAN_KEY, None)
            if parent_span is not None:
                # Set final attributes and events using common method
                _setup_span_attributes_and_events(parent_span, final_attributes, final_events)
                
                parent_span.end()
        if token is not None:
            detach(token)
    except Exception as e:
        logger.warning(f"Failed to stop trace: {e}")

def start_scope(
    scope_name: str, 
    scope_value: Optional[str] = None
) -> object:
    """
    Start a new scope with the given name and optional value. If no value is provided, a random UUID will be generated.
    All the spans, across traces created after this call will have the scope attached until the scope is stopped.
    
    Args:
        scope_name: The name of the scope.
        scope_value: Optional value of the scope. If None, a random UUID will be generated.
    
    Returns:
        Token: A token representing the attached context for the scope. This token is to be used later to stop the current scope.
    """
    try:
        # Set the scope using existing utility
        token = set_scope(scope_name, scope_value)
        return token
    except Exception as e:
        logger.warning(f"Failed to start scope: {e}")
        return None

def stop_scope(
    token: object
) -> None:
    """
    Stop the active scope. All the spans created after this will not have the scope attached.
    
    Args:
        token: The token that was returned when the scope was started.
    
    Returns:
        None
    """
    try:
        # Remove the scope
        remove_scope(token)
    except Exception as e:
        logger.warning(f"Failed to stop scope: {e}")
    return


@contextmanager
def monocle_trace(
    span_name: Optional[str] = None,
    attributes: Optional[Dict[str, Any]] = None,
    events: Optional[List[Dict[str, Any]]] = None
):
    """
    Context manager to start and stop a trace. All the spans, across traces created within the encapsulated code will have same trace ID
    
    Args:
        span_name: Optional custom span name.
        attributes: Optional dictionary of custom attributes to set on the span.
        events: Optional list of events to add to the span at start.
    """
    try:
        tracer = get_tracer(instrumenting_module_name=MONOCLE_INSTRUMENTOR, tracer_provider=get_tracer_provider())
        span_name = span_name or "custom_span"
                
        with tracer.start_as_current_span(span_name) as span:
            # Set default monocle attributes
            SpanHandler.set_default_monocle_attributes(span)
            if SpanHandler.is_root_span(span):
                SpanHandler.set_workflow_properties(span)
            
            # Set custom attributes and events using common method
            _setup_span_attributes_and_events(span, attributes, events)
            
            try:
                yield
            finally:
                pass

            
    except Exception as e:
        logger.warning(f"Failed in monocle_trace: {e}")
        yield  # Still yield to not break the context manager

@asynccontextmanager
async def amonocle_trace(
    span_name: Optional[str] = None,
    attributes: Optional[Dict[str, Any]] = None,
    events: Optional[List[Dict[str, Any]]] = None
):
    """
    Async context manager to start and stop a trace. All the spans, across traces created within the encapsulated code will have same trace ID
    
    Args:
        span_name: Optional custom span name.
        attributes: Optional dictionary of custom attributes to set on the span.
        events: Optional list of events to add to the span at start.
    """
    try:
        tracer = get_tracer(instrumenting_module_name=MONOCLE_INSTRUMENTOR, tracer_provider=get_tracer_provider())
        span_name = span_name or "custom_span"

        with tracer.start_as_current_span(span_name) as span:
            # Set default monocle attributes
            SpanHandler.set_default_monocle_attributes(span)
            if SpanHandler.is_root_span(span):
                SpanHandler.set_workflow_properties(span)
            
            # Set custom attributes and events using common method
            _setup_span_attributes_and_events(span, attributes, events)
            
            try:
                yield
            finally:
                pass
                
            
    except Exception as e:
        logger.warning(f"Failed in amonocle_trace: {e}")
        yield  # Still yield to not break the context manager

@contextmanager
def monocle_trace_scope(
    scope_name: str, 
    scope_value: Optional[str] = None
):
    """
    Context manager to start and stop a scope. All the spans, across traces created within the encapsulated code will have the scope attached.
    
    Args:
        scope_name: The name of the scope.
        scope_value: Optional value of the scope. If None, a random UUID will be generated.
    """
    token = start_scope(scope_name, scope_value)
    try:
        yield
    finally:
        stop_scope(token)

@asynccontextmanager
async def amonocle_trace_scope(
    scope_name: str, 
    scope_value: Optional[str] = None
):
    """
    Async context manager to start and stop a scope. All the spans, across traces created within the encapsulated code will have the scope attached.
    
    Args:
        scope_name: The name of the scope.
        scope_value: Optional value of the scope. If None, a random UUID will be generated.
    """
    token = start_scope(scope_name, scope_value)
    try:
        yield
    finally:
        stop_scope(token)
    
def monocle_trace_scope_method(
    scope_name: str, 
    scope_value: Optional[str] = None
):
    """
    Decorator to start and stop a scope for a method. All the spans, across traces created in the method will have the scope attached.
    
    Args:
        scope_name: The name of the scope.
        scope_value: Optional value of the scope. If None, a random UUID will be generated.
    """
    def decorator(func):
        if inspect.iscoroutinefunction(func):
            @wraps(func)
            async def wrapper(*args, **kwargs):
                async with amonocle_trace_scope(
                    scope_name, scope_value
                ):
                    result = await func(*args, **kwargs)
                    return result
            return wrapper
        else:
            @wraps(func)
            def wrapper(*args, **kwargs):
                with monocle_trace_scope(
                    scope_name, scope_value
                ):
                    result = func(*args, **kwargs)
                    return result
            return wrapper
    return decorator

def monocle_trace_method(
    span_name: Optional[str] = None
):
    """
    Decorator to start and stop a trace for a method. All the spans created in the method will be part of the same trace.
    
    Args:
        span_name: Optional custom span name. If None, uses the decorated function's name.
    """
    
    def decorator(func):
        tracer = get_tracer(instrumenting_module_name=MONOCLE_INSTRUMENTOR, tracer_provider=get_tracer_provider())
        handler = SpanHandler()
        source_path= func.__code__.co_filename + ":" + str(func.__code__.co_firstlineno)
        # Use function name as span name if not provided
        effective_span_name = span_name or func.__name__ or "custom_span"

        if inspect.iscoroutinefunction(func):
            @wraps(func)
            async def wrapper(*args, **kwargs):
                return await atask_wrapper(
                    tracer=tracer,
                    handler=handler,
                    to_wrap={
                        "span_name": effective_span_name,
                        "output_processor":{
                            "type": "custom",
                        }
                    }
                )(  wrapped=func,                        
                    instance=None,
                    source_path=source_path,
                    args=args,
                    kwargs=kwargs)
            return wrapper
        else:
            @wraps(func)
            def wrapper(*args, **kwargs):
                return task_wrapper(
                    tracer=tracer,
                    handler=handler,
                    to_wrap={
                        "span_name": effective_span_name,
                        "output_processor":{
                            "type": "custom",
                        }
                    }
                )(  wrapped=func,                        
                    instance=None,
                    source_path=source_path,
                    args=args,
                    kwargs=kwargs)
            return wrapper
    return decorator

def monocle_trace_http_route(func):
    """
    Decorator to start and stop a continue traces and scope for a http route. It will also initiate new scopes from the http headers if configured in ``monocle_scopes.json``
    All the spans, across traces created in the route will have the scope attached.
    """
    if inspect.iscoroutinefunction(func):
        @wraps(func)
        async def wrapper(*args, **kwargs):
            return await http_async_route_handler(func, *args, **kwargs)
        return wrapper
    else:
        @wraps(func)
        def wrapper(*args, **kwargs):
            return http_route_handler(func, *args, **kwargs)
        return wrapper


def _setup_span_attributes_and_events(
    span,
    attributes: Optional[Dict[str, Any]] = None,
    events: Optional[List[Dict[str, Any]]] = None
) -> None:
    """
    Common method to set attributes and events on a span.
    
    Args:
        span: The span to configure
        attributes: Optional dictionary of custom attributes to set on the span
        events: Optional list of events to add to the span
    """
    # Set custom attributes if provided
    if attributes:
        for key, value in attributes.items():
            if value is not None:
                span.set_attribute(key, value)
    
    # Add custom events if provided
    if events:
        for event in events:
            event_name = event.get('name')
            event_attributes = event.get('attributes', {})
            if event_name:
                span.add_event(event_name, event_attributes)




"""
This module provides utility functions for extracting system, user,
and assistant messages from various input formats.
"""

import logging
from monocle_apptrace.instrumentation.common.utils import (
    Option,
    get_json_dumps,
    try_option,
    get_exception_message,
    get_parent_span,
    get_status_code,
)
from monocle_apptrace.instrumentation.common.span_handler import NonFrameworkSpanHandler, WORKFLOW_TYPE_MAP
from monocle_apptrace.instrumentation.metamodel.finish_types import (
    map_openai_finish_reason_to_finish_type,
    OPENAI_FINISH_REASON_MAPPING
)
from monocle_apptrace.instrumentation.common.constants import CHILD_ERROR_CODE

logger = logging.getLogger(__name__)

def extract_messages(kwargs):
    """Extract system and user messages"""
    try:
        messages = []
        if 'instructions' in kwargs:
            messages.append({'system': kwargs.get('instructions', {})})
        if 'input' in kwargs:
            if isinstance(kwargs['input'], str): 
                messages.append({'user': kwargs.get('input', "")})
            # [
            #     {
            #         "role": "developer",
            #         "content": "Talk like a pirate."
            #     },
            #     {
            #         "role": "user",
            #         "content": "Are semicolons optional in JavaScript?"
            #     }
            # ]
            if isinstance(kwargs['input'], list):
                for item in kwargs['input']:
                    if isinstance(item, dict) and 'role' in item and 'content' in item:
                        messages.append({item['role']: item['content']})
        if 'messages' in kwargs and len(kwargs['messages']) >0:
            for msg in kwargs['messages']:
                if msg.get('content') and msg.get('role'):
                    messages.append({msg['role']: msg['content']})

        return [get_json_dumps(message) for message in messages]
    except Exception as e:
        logger.warning("Warning: Error occurred in extract_messages: %s", str(e))
        return []


def extract_assistant_message(arguments):
    try:
        messages = []
        status = get_status_code(arguments)
        if status == 'success' or status == 'completed':
            response = arguments["result"]
            if hasattr(response, "output_text") and len(response.output_text):
                role = response.role if hasattr(response, "role") else "assistant"
                messages.append({role: response.output_text})
            if (
                response is not None
                and hasattr(response, "choices")
                and len(response.choices) > 0
            ):
                if hasattr(response.choices[0], "message"):
                    role = (
                        response.choices[0].message.role
                        if hasattr(response.choices[0].message, "role")
                        else "assistant"
                    )
                    messages.append({role: response.choices[0].message.content})
            return get_json_dumps(messages[0]) if messages else ""
        else:
            if arguments["exception"] is not None:
                return get_exception_message(arguments)
            elif hasattr(arguments["result"], "error"):
                return arguments["result"].error
        
    except (IndexError, AttributeError) as e:
        logger.warning(
            "Warning: Error occurred in extract_assistant_message: %s", str(e)
        )
        return None


def extract_provider_name(instance):
    provider_url: Option[str] = try_option(getattr, instance._client.base_url, 'host')
    return provider_url.unwrap_or(None)


def extract_inference_endpoint(instance):
    inference_endpoint: Option[str] = try_option(getattr, instance._client, 'base_url').map(str)
    if inference_endpoint.is_none() and "meta" in instance.client.__dict__:
        inference_endpoint = try_option(getattr, instance.client.meta, 'endpoint_url').map(str)

    return inference_endpoint.unwrap_or(extract_provider_name(instance))

def resolve_from_alias(my_map, alias):
    """Find a alias that is not none from list of aliases"""

    for i in alias:
        if i in my_map.keys():
            return my_map[i]
    return None


def update_input_span_events(kwargs):
    if 'input' in kwargs and isinstance(kwargs['input'], list):
        query = ' '.join(kwargs['input'])
        return query


def update_output_span_events(results):
    if hasattr(results,'data') and isinstance(results.data, list):
        embeddings = results.data
        embedding_strings = [f"index={e.index}, embedding={e.embedding}" for e in embeddings]
        output = '\n'.join(embedding_strings)
        if len(output) > 100:
            output = output[:100] + "..."
        return output


def update_span_from_llm_response(response):
    meta_dict = {}
    if response is not None and hasattr(response, "usage"):
        if hasattr(response, "usage") and response.usage is not None:
            token_usage = response.usage
        else:
            response_metadata = response.response_metadata
            token_usage = response_metadata.get("token_usage")
        if token_usage is not None:
            meta_dict.update({"completion_tokens": getattr(token_usage,"completion_tokens",None) or getattr(token_usage,"output_tokens",None)})
            meta_dict.update({"prompt_tokens": getattr(token_usage, "prompt_tokens", None) or getattr(token_usage, "input_tokens", None)})
            meta_dict.update({"total_tokens": getattr(token_usage,"total_tokens")})
    return meta_dict

def extract_vector_input(vector_input: dict):
    if 'input' in vector_input:
        return vector_input['input']
    return ""

def extract_vector_output(vector_output):
    try:
        if hasattr(vector_output, 'data') and len(vector_output.data) > 0:
            return vector_output.data[0].embedding
    except Exception as e:
        pass
    return ""

def get_inference_type(instance):
    inference_type: Option[str] = try_option(getattr, instance._client, '_api_version')
    if inference_type.unwrap_or(None):
        return 'azure_openai'
    else:
        return 'openai'

class OpenAISpanHandler(NonFrameworkSpanHandler):
    def is_teams_span_in_progress(self) -> bool:
        return self.is_framework_span_in_progress() and self.get_workflow_name_in_progress() == WORKFLOW_TYPE_MAP["teams.ai"]

    # If openAI is being called by Teams AI SDK, then retain the metadata part of the span events
    def skip_processor(self, to_wrap, wrapped, instance, span, args, kwargs) -> list[str]:
        if self.is_teams_span_in_progress():
            return ["attributes", "events.data.input", "events.data.output"]
        else:
            return super().skip_processor(to_wrap, wrapped, instance, span, args, kwargs)

    def hydrate_events(self, to_wrap, wrapped, instance, args, kwargs, ret_result, span, parent_span=None, ex:Exception=None) -> bool:
        # If openAI is being called by Teams AI SDK, then copy parent
        if self.is_teams_span_in_progress() and ex is None:
            return super().hydrate_events(to_wrap, wrapped, instance, args, kwargs, ret_result, span=parent_span, parent_span=None, ex=ex)

        return super().hydrate_events(to_wrap, wrapped, instance, args, kwargs, ret_result, span, parent_span=parent_span, ex=ex)

    def post_task_processing(self, to_wrap, wrapped, instance, args, kwargs, result, ex, span, parent_span):
        # TeamsAI doesn't capture the status and other metadata from underlying OpenAI SDK.
        # Thus we save the OpenAI status code in the parent span and retrieve it here to preserve meaningful error codes.
        if self.is_teams_span_in_progress() and ex is not None:
            if len(span.events) > 1 and span.events[1].name == "data.output" and span.events[1].attributes.get("error_code") is not None:
                parent_span.set_attribute(CHILD_ERROR_CODE, span.events[1].attributes.get("error_code"))
        super().post_task_processing(to_wrap, wrapped, instance, args, kwargs, result, ex, span, parent_span)

def extract_finish_reason(arguments):
    """Extract finish_reason from OpenAI response"""
    try:
        if arguments["exception"] is not None:
            if hasattr(arguments["exception"], "code") and arguments["exception"].code in OPENAI_FINISH_REASON_MAPPING.keys():
                return arguments["exception"].code
        response = arguments["result"]
        
        # Handle streaming responses
        if hasattr(response, "finish_reason") and response.finish_reason:
            return response.finish_reason
            
        # Handle non-streaming responses
        if response is not None and hasattr(response, "choices") and len(response.choices) > 0:
            if hasattr(response.choices[0], "finish_reason"):
                return response.choices[0].finish_reason
    except (IndexError, AttributeError) as e:
        logger.warning("Warning: Error occurred in extract_finish_reason: %s", str(e))
        return None
    return None

def map_finish_reason_to_finish_type(finish_reason):
    """Map OpenAI finish_reason to finish_type based on the possible errors mapping"""
    return map_openai_finish_reason_to_finish_type(finish_reason)



from opentelemetry.context import set_value, attach, detach
from monocle_apptrace.instrumentation.common.span_handler import SpanHandler
from monocle_apptrace.instrumentation.metamodel.mcp._helper import (
   get_name

)
from monocle_apptrace.instrumentation.metamodel.langgraph.entities.inference import (
    AGENT_DELEGATION, AGENT_REQUEST
)

class MCPAgentHandler(SpanHandler):
    def skip_span(self, to_wrap, wrapped, instance, args, kwargs) -> bool:
        return get_name({"args": args, "kwargs": kwargs}) is None or args[0].root.method == "tools/list"



from monocle_apptrace.instrumentation.common.wrapper import atask_wrapper, task_wrapper
from monocle_apptrace.instrumentation.metamodel.azfunc.entities.http import AZFUNC_HTTP_PROCESSOR

AZFUNC_HTTP_METHODS = [
    {
        "package": "monocle_apptrace.instrumentation.metamodel.azfunc.wrapper",
        "object": "AzureFunctionRouteWrapper",
        "method": "run_async",
        "span_name": "azure_function_route",
        "wrapper_method": atask_wrapper,
        "span_handler": "azure_func_handler",
        "output_processor": AZFUNC_HTTP_PROCESSOR
    },
    {
        "package": "monocle_apptrace.instrumentation.metamodel.azfunc.wrapper",
        "object": "AzureFunctionRouteWrapper",
        "method": "run_sync",
        "span_name": "azure_function_route",
        "wrapper_method": task_wrapper,
        "span_handler": "azure_func_handler",
        "output_processor": AZFUNC_HTTP_PROCESSOR
    }
]



from monocle_apptrace.instrumentation.common.wrapper import task_wrapper
from monocle_apptrace.instrumentation.metamodel.gemini.entities.inference import (
    INFERENCE,
)
from monocle_apptrace.instrumentation.metamodel.gemini.entities.retrieval import (
    RETRIEVAL,
)

GEMINI_METHODS = [
    {
      "package": "google.genai.models",
      "object": "Models",
      "method": "generate_content",
      "wrapper_method": task_wrapper,
      "output_processor": INFERENCE,
    },
    {
      "package": "google.genai.models",
      "object": "Models",
      "method": "embed_content",
      "wrapper_method": task_wrapper,
      "output_processor": RETRIEVAL,
    }
]


#pylint: disable=consider-using-with

from os import linesep, path
from io import TextIOWrapper
from datetime import datetime
from typing import Optional, Callable, Sequence, Dict, Tuple
from opentelemetry.sdk.trace import ReadableSpan
from opentelemetry.sdk.trace.export import SpanExporter, SpanExportResult
from opentelemetry.sdk.resources import SERVICE_NAME
from monocle_apptrace.exporters.base_exporter import SpanExporterBase
from monocle_apptrace.exporters.exporter_processor import ExportTaskProcessor

DEFAULT_FILE_PREFIX:str = "monocle_trace_"
DEFAULT_TIME_FORMAT:str = "%Y-%m-%d_%H.%M.%S"
HANDLE_TIMEOUT_SECONDS: int = 60  # 1 minute timeout

class FileSpanExporter(SpanExporterBase):
    def __init__(
        self,
        service_name: Optional[str] = None,
        out_path:str = ".",
        file_prefix = DEFAULT_FILE_PREFIX,
        time_format = DEFAULT_TIME_FORMAT,
        formatter: Callable[
            [ReadableSpan], str
        ] = lambda span: span.to_json()
        + linesep,
        task_processor: Optional[ExportTaskProcessor] = None
    ):
        super().__init__()
        # Dictionary to store file handles: {trace_id: (file_handle, file_path, creation_time, first_span)}
        self.file_handles: Dict[int, Tuple[TextIOWrapper, str, datetime, bool]] = {}
        self.formatter = formatter
        self.service_name = service_name
        self.output_path = out_path
        self.file_prefix = file_prefix
        self.time_format = time_format
        self.task_processor = task_processor
        self.is_first_span_in_file = True  # Track if this is the first span in the current file
        if self.task_processor is not None:
            self.task_processor.start()

    def export(self, spans: Sequence[ReadableSpan]) -> SpanExportResult:
        is_root_span = any(not span.parent for span in spans)
        if self.task_processor is not None and callable(getattr(self.task_processor, 'queue_task', None)):
            # Check if any span is a root span (no parent)
            self.task_processor.queue_task(self._process_spans, spans, is_root_span)
            return SpanExportResult.SUCCESS
        else:
            return self._process_spans(spans, is_root_span=is_root_span)

    def _cleanup_expired_handles(self) -> None:
        """Close and remove file handles that have exceeded the timeout."""
        current_time = datetime.now()
        expired_trace_ids = []
        
        for trace_id, (handle, file_path, creation_time, _) in self.file_handles.items():
            if (current_time - creation_time).total_seconds() > HANDLE_TIMEOUT_SECONDS:
                expired_trace_ids.append(trace_id)
        
        for trace_id in expired_trace_ids:
            self._close_trace_handle(trace_id)

    def _get_or_create_handle(self, trace_id: int, service_name: str) -> Tuple[TextIOWrapper, str, bool]:
        """Get existing handle or create new one for the trace_id."""
        self._cleanup_expired_handles()
        
        if trace_id in self.file_handles:
            handle, file_path, creation_time, first_span = self.file_handles[trace_id]
            return handle, file_path, first_span
        
        # Create new handle
        file_path = path.join(self.output_path,
                             self.file_prefix + service_name + "_" + hex(trace_id) + "_"
                             + datetime.now().strftime(self.time_format) + ".json")
        
        try:
            handle = open(file_path, "w", encoding='UTF-8')
            handle.write("[")
            self.file_handles[trace_id] = (handle, file_path, datetime.now(), True)
            return handle, file_path, True
        except Exception as e:
            print(f"Error creating file {file_path}: {e}")
            return None, file_path, True

    def _close_trace_handle(self, trace_id: int) -> None:
        """Close and remove a specific trace handle."""
        if trace_id in self.file_handles:
            handle, file_path, creation_time, _ = self.file_handles[trace_id]
            try:
                if handle is not None:
                    handle.write("]")
                    handle.close()
            except Exception as e:
                print(f"Error closing file {file_path}: {e}")
            finally:
                del self.file_handles[trace_id]

    def _mark_span_written(self, trace_id: int) -> None:
        """Mark that a span has been written for this trace (no longer first span)."""
        if trace_id in self.file_handles:
            handle, file_path, creation_time, _ = self.file_handles[trace_id]
            self.file_handles[trace_id] = (handle, file_path, creation_time, False)

    def _process_spans(self, spans: Sequence[ReadableSpan], is_root_span: bool = False) -> SpanExportResult:
        # Group spans by trace_id for efficient processing
        spans_by_trace = {}
        root_span_traces = set()
        
        for span in spans:
            if self.skip_export(span):
                continue
            
            trace_id = span.context.trace_id
            if trace_id not in spans_by_trace:
                spans_by_trace[trace_id] = []
            spans_by_trace[trace_id].append(span)
            
            # Check if this span is a root span
            if not span.parent:
                root_span_traces.add(trace_id)
        
        # Process spans for each trace
        for trace_id, trace_spans in spans_by_trace.items():
            service_name = trace_spans[0].resource.attributes.get(SERVICE_NAME, "unknown")
            handle, file_path, is_first_span = self._get_or_create_handle(trace_id, service_name)
            
            if handle is None:
                continue
            
            for span in trace_spans:
                if not is_first_span:
                    try:
                        handle.write(",")
                    except Exception as e:
                        print(f"Error writing comma to file {file_path} for span {span.context.span_id}: {e}")
                        continue
                
                try:
                    handle.write(self.formatter(span))
                    if is_first_span:
                        self._mark_span_written(trace_id)
                        is_first_span = False
                except Exception as e:
                    print(f"Error formatting span {span.context.span_id}: {e}")
                    continue
        
        # Close handles for traces with root spans
        for trace_id in root_span_traces:
            self._close_trace_handle(trace_id)
        
        # Flush remaining handles
        for trace_id, (handle, file_path, _, _) in self.file_handles.items():
            if trace_id not in root_span_traces:
                try:
                    if handle is not None:
                        handle.flush()
                except Exception as e:
                    print(f"Error flushing file {file_path}: {e}")
        
        return SpanExportResult.SUCCESS

    def force_flush(self, timeout_millis: int = 30000) -> bool:
        """Flush all open file handles."""
        for trace_id, (handle, file_path, _, _) in self.file_handles.items():
            try:
                if handle is not None:
                    handle.flush()
            except Exception as e:
                print(f"Error flushing file {file_path}: {e}")
        return True

    def shutdown(self) -> None:
        """Close all file handles and stop task processor."""
        if hasattr(self, 'task_processor') and self.task_processor is not None:
            self.task_processor.stop()
        
        # Close all remaining file handles
        trace_ids_to_close = list(self.file_handles.keys())
        for trace_id in trace_ids_to_close:
            self._close_trace_handle(trace_id)



from monocle_apptrace.instrumentation.common.wrapper import (
    ascopes_wrapper,
    atask_wrapper,
    task_wrapper,
    ascope_wrapper,
)
from monocle_apptrace.instrumentation.metamodel.teamsai.entities.inference.teamsai_output_processor import (
    TEAMAI_OUTPUT_PROCESSOR,
)
from monocle_apptrace.instrumentation.metamodel.teamsai.entities.inference.actionplanner_output_processor import (
    ACTIONPLANNER_OUTPUT_PROCESSOR,
) 
def get_id(args, kwargs):
    """
    Extracts the ID from the context.
    """
    scopes: dict[str, dict[str:str]] = {}
    context = kwargs.get("context")
    if context and context.activity and context.activity.channel_id:
        channel_id = context.activity.channel_id or ""
        scopes[f"teams.channel.channel_id"] = channel_id
        if channel_id == "msteams":
            scopes[f"msteams.activity.type"] = context.activity.type or ""

            if hasattr(context.activity,"conversation"):
                scopes[f"msteams.conversation.id"] = context.activity.conversation.id or ""
                scopes[f"msteams.conversation.type"] = context.activity.conversation.conversation_type or ""
                scopes[f"msteams.conversation.name"] = context.activity.conversation.name or ""

            if hasattr(context.activity,"from_property"):
                scopes[f"msteams.user.from_property.id"] = context.activity.from_property.id or ""
                scopes[f"msteams.user.from_property.name"] = context.activity.from_property.name or ""
                scopes[f"msteams.user.from_property.role"] = context.activity.from_property.role or ""

            if hasattr(context.activity,"recipient"):
                scopes[f"msteams.recipient.id"] = context.activity.recipient.id or ""

            if hasattr(context.activity,"channel_data"):
                if "tenant" in context.activity.channel_data:
                    scopes[f"msteams.channel_data.tenant.id"] = context.activity.channel_data['tenant']['id'] or ""
                if "team" in context.activity.channel_data:
                    scopes[f"msteams.channel_data.team.id"] = context.activity.channel_data['team']['id'] or ""
                    if "name" in context.activity.channel_data['team']:
                        scopes[f"msteams.channel_data.team.name"] = context.activity.channel_data['team']['name'] or ""
                if "channel" in context.activity.channel_data:
                    scopes[f"msteams.channel_data.channel.id"] = context.activity.channel_data['channel']['id'] or ""
                    if "name" in context.activity.channel_data['channel']:
                        scopes[f"msteams.channel_data.channel.name"] = context.activity.channel_data['channel']['name'] or ""
    return scopes

TEAMAI_METHODS = [
    {
        "package": "teams.ai.models.openai_model",
        "object": "OpenAIModel",
        "method": "complete_prompt",
        "wrapper_method": atask_wrapper,
        "output_processor": TEAMAI_OUTPUT_PROCESSOR,
    },
    {
        "package": "teams.ai.planners.action_planner",
        "object": "ActionPlanner",
        "method": "complete_prompt",
        "wrapper_method": atask_wrapper,
        "output_processor": ACTIONPLANNER_OUTPUT_PROCESSOR,
    },
    {
        "package": "teams.ai.planners.action_planner",
        "object": "ActionPlanner",
        "method": "complete_prompt",
        "scope_values": get_id,
        "wrapper_method": ascopes_wrapper,  
    }
]



from monocle_apptrace.instrumentation.common.wrapper import atask_wrapper, task_wrapper
from monocle_apptrace.instrumentation.metamodel.openai.entities.inference import (
    INFERENCE,
)
from monocle_apptrace.instrumentation.metamodel.openai.entities.retrieval import (
    RETRIEVAL,
)

OPENAI_METHODS = [
    {
        "package": "openai.resources.chat.completions",
        "object": "Completions",
        "method": "create",
        "wrapper_method": task_wrapper,
        "span_handler": "openai_handler",
        "output_processor": INFERENCE
    },
    {
        "package": "openai.resources.chat.completions",
        "object": "AsyncCompletions",
        "method": "create",
        "wrapper_method": atask_wrapper,
        "span_handler": "openai_handler",
        "output_processor": INFERENCE
    },
    {
        "package": "openai.resources.embeddings",
        "object": "Embeddings",
        "method": "create",
        "wrapper_method": task_wrapper,
        "span_handler": "openai_handler",
        "output_processor": RETRIEVAL
    },
    {
        "package": "openai.resources.embeddings",
        "object": "AsyncEmbeddings",
        "method": "create",
        "wrapper_method": atask_wrapper,
        "span_handler": "openai_handler",
        "output_processor": RETRIEVAL
    },
    {
        "package": "openai.resources.responses",
        "object": "Responses",
        "method": "create",
        "wrapper_method": task_wrapper,
        "span_handler": "openai_handler",
        "output_processor": INFERENCE
    },
    {
        "package": "openai.resources.responses",
        "object": "AsyncResponses",
        "method": "create",
        "wrapper_method": atask_wrapper,
        "span_handler": "openai_handler",
        "output_processor": INFERENCE
    }

]



import logging
import inspect
from typing import Collection, Dict, List, Union
import uuid
import inspect
from opentelemetry import trace
from opentelemetry.context import attach, get_value, set_value
from opentelemetry.instrumentation.instrumentor import BaseInstrumentor
from opentelemetry.instrumentation.utils import unwrap
from opentelemetry.sdk.trace import TracerProvider, Span
from opentelemetry.sdk.resources import SERVICE_NAME, Resource
from opentelemetry.sdk.trace import Span, TracerProvider
from opentelemetry.sdk.trace.export import BatchSpanProcessor, SpanProcessor
from opentelemetry.sdk.trace.export import SpanExporter
from opentelemetry.trace import get_tracer
from wrapt import wrap_function_wrapper
from monocle_apptrace.exporters.monocle_exporters import get_monocle_exporter
from monocle_apptrace.instrumentation.common.span_handler import SpanHandler, NonFrameworkSpanHandler
from monocle_apptrace.instrumentation.common.wrapper_method import (
    DEFAULT_METHODS_LIST,
    WrapperMethod,
    MONOCLE_SPAN_HANDLERS
)
from monocle_apptrace.instrumentation.common.wrapper import scope_wrapper, ascope_wrapper, monocle_wrapper, amonocle_wrapper
from monocle_apptrace.instrumentation.common.utils import (
    load_scopes
)
from monocle_apptrace.instrumentation.common.constants import MONOCLE_INSTRUMENTOR
from functools import wraps

logger = logging.getLogger(__name__)

SESSION_PROPERTIES_KEY = "session"

_instruments = ()

monocle_tracer_provider: TracerProvider = None

class MonocleInstrumentor(BaseInstrumentor):
    workflow_name: str = ""
    user_wrapper_methods: list[Union[dict,WrapperMethod]] = [],
    exporters: list[SpanExporter] = [],
    instrumented_method_list: list[object] = []
    handlers:Dict[str,SpanHandler] = None # dict of handlers
    union_with_default_methods: bool = False

    def __init__(
            self,
            handlers,
            user_wrapper_methods: list[Union[dict,WrapperMethod]] = None,
            exporters: list[SpanExporter] = None,
            union_with_default_methods: bool = True
            ) -> None:
        self.user_wrapper_methods = user_wrapper_methods or []
        self.handlers = handlers
        self.exporters = exporters
        if self.handlers is not None:
            for key, val in MONOCLE_SPAN_HANDLERS.items():
                if key not in self.handlers:
                    self.handlers[key] = val
        else:
            self.handlers = MONOCLE_SPAN_HANDLERS
        self.union_with_default_methods = union_with_default_methods
        super().__init__()

    def get_instrumentor(self, tracer):
        def instrumented_endpoint_invoke(to_wrap,wrapped, span_name, instance,fn):
            if inspect.iscoroutinefunction(fn):
                @wraps(fn)
                async def with_instrumentation(*args, **kwargs):
                    boto_method_to_wrap = to_wrap.copy()
                    boto_method_to_wrap['skip_span'] = False
                    return await amonocle_wrapper(tracer, NonFrameworkSpanHandler(),
                            boto_method_to_wrap, fn, instance, "", args, kwargs)
            else:
                @wraps(fn)
                def with_instrumentation(*args, **kwargs):
                    boto_method_to_wrap = to_wrap.copy()
                    boto_method_to_wrap['skip_span'] = False
                    return monocle_wrapper(tracer, NonFrameworkSpanHandler(),
                            boto_method_to_wrap, fn, instance, "", args, kwargs)
            return with_instrumentation
        return instrumented_endpoint_invoke

    def instrumentation_dependencies(self) -> Collection[str]:
        return _instruments

    def _instrument(self, **kwargs):
        tracer_provider: TracerProvider = kwargs.get("tracer_provider")
        set_tracer_provider(tracer_provider)
        tracer = get_tracer(instrumenting_module_name=MONOCLE_INSTRUMENTOR, tracer_provider=tracer_provider)

        final_method_list = []
        if self.union_with_default_methods is True:
            final_method_list= final_method_list + DEFAULT_METHODS_LIST

        for method in self.user_wrapper_methods:
            if isinstance(method, dict):
                final_method_list.append(method)
            elif isinstance(method, WrapperMethod):
                final_method_list.append(method.to_dict())

        for method in load_scopes():
            if method.get('async', False):
                method['wrapper_method'] = ascope_wrapper
            else:
                method['wrapper_method'] = scope_wrapper
            final_method_list.append(method)
        
        for method_config in final_method_list:
            target_package = method_config.get("package", None)
            target_object = method_config.get("object", None)
            target_method = method_config.get("method", None)
            wrapped_by = method_config.get("wrapper_method", None)
            #get the requisite handler or default one
            handler_key = method_config.get("span_handler",'default')
            try:
                handler =  self.handlers.get(handler_key)
                if not handler:
                    logger.warning("incorrect or empty handler falling back to default handler")
                    handler = self.handlers.get('default')
                handler.set_instrumentor(self.get_instrumentor(tracer))
                wrap_function_wrapper(
                    target_package,
                    f"{target_object}.{target_method}" if target_object else target_method,
                    wrapped_by(tracer, handler, method_config),
                )
                self.instrumented_method_list.append(method_config)
            except ModuleNotFoundError as e:
                logger.debug(f"ignoring module {e.name}")

            except Exception as ex:
                logger.error(f"""_instrument wrap exception: {str(ex)}
                            for package: {target_package},
                            object:{target_object},
                            method:{target_method}""")

    def _uninstrument(self, **kwargs):
        for wrapped_method in self.instrumented_method_list:
            try:
                wrap_package = wrapped_method.get("package")
                wrap_object = wrapped_method.get("object")
                wrap_method = wrapped_method.get("method")
                unwrap(
                    f"{wrap_package}.{wrap_object}" if wrap_object else wrap_package,
                    wrap_method,
                )
            except Exception as ex:
                logger.error(f"""_instrument unwrap exception: {str(ex)}
                             for package: {wrap_package},
                             object:{wrap_object},
                             method:{wrap_method}""")

def set_tracer_provider(tracer_provider: TracerProvider):
    global monocle_tracer_provider
    monocle_tracer_provider = tracer_provider

def get_tracer_provider() -> TracerProvider:
    global monocle_tracer_provider
    return monocle_tracer_provider

def setup_monocle_telemetry(
        workflow_name: str,
        span_processors: List[SpanProcessor] = None,
        span_handlers: Dict[str,SpanHandler] = None,
        wrapper_methods: List[Union[dict,WrapperMethod]] = None,
        union_with_default_methods: bool = True,
        monocle_exporters_list:str = None) -> None:
    """
    Set up Monocle telemetry for the application.

    Parameters
    ----------
    workflow_name : str
        The name of the workflow to be used as the service name in telemetry.
    span_processors : List[SpanProcessor], optional
        Custom span processors to use instead of the default ones. If None, 
        BatchSpanProcessors with Monocle exporters will be used. This can't be combined with `monocle_exporters_list`.
    span_handlers : Dict[str, SpanHandler], optional
        Dictionary of span handlers to be used by the instrumentor, mapping handler names to handler objects.
    wrapper_methods : List[Union[dict, WrapperMethod]], optional
        Custom wrapper methods for instrumentation. If None, default methods will be used.
    union_with_default_methods : bool, default=True
        If True, combine the provided wrapper_methods with the default methods.
        If False, only use the provided wrapper_methods.
    monocle_exporters_list : str, optional
        Comma-separated list of exporters to use. This will override the env setting MONOCLE_EXPORTERS.
        Supported exporters are: s3, blob, okahu, file, memory, console. This can't be combined with `span_processors`.
    """
    resource = Resource(attributes={
        SERVICE_NAME: workflow_name
    })
    if span_processors and monocle_exporters_list:
        raise ValueError("span_processors and monocle_exporters_list can't be used together")
    exporters:List[SpanExporter] = get_monocle_exporter(monocle_exporters_list)
    span_processors = span_processors or [BatchSpanProcessor(exporter) for exporter in exporters]
    set_tracer_provider(TracerProvider(resource=resource))
    attach(set_value("workflow_name", workflow_name))
    tracer_provider_default = trace.get_tracer_provider()
    provider_type = type(tracer_provider_default).__name__
    is_proxy_provider = "Proxy" in provider_type
    for processor in span_processors:
        processor.on_start = on_processor_start
        if not is_proxy_provider:
            tracer_provider_default.add_span_processor(processor)
        else:
            get_tracer_provider().add_span_processor(processor)
    if is_proxy_provider:
        trace.set_tracer_provider(get_tracer_provider())
    instrumentor = MonocleInstrumentor(user_wrapper_methods=wrapper_methods or [], exporters=exporters,
                                       handlers=span_handlers, union_with_default_methods = union_with_default_methods)
    # instrumentor.app_name = workflow_name
    if not instrumentor.is_instrumented_by_opentelemetry:
        instrumentor.instrument(trace_provider=get_tracer_provider())

    return instrumentor

def on_processor_start(span: Span, parent_context):
    context_properties = get_value(SESSION_PROPERTIES_KEY)
    if context_properties is not None:
        for key, value in context_properties.items():
            span.set_attribute(
                f"{SESSION_PROPERTIES_KEY}.{key}", value
            )

def set_context_properties(properties: dict) -> None:
    attach(set_value(SESSION_PROPERTIES_KEY, properties))

def is_valid_trace_id_uuid(traceId: str) -> bool:
    try:
        uuid.UUID(traceId)
        return True
    except:
        pass
    return False

from monocle_apptrace.instrumentation.common.method_wrappers import (
    monocle_trace_scope_method,
    amonocle_trace_scope,
    monocle_trace,
    amonocle_trace,
    monocle_trace_method,
    monocle_trace_http_route,
    start_scope,
    stop_scope,
    start_trace,
    stop_trace,
    http_route_handler,
    monocle_trace_scope,
)




from .instrumentor import (
    setup_monocle_telemetry, 
    start_trace, 
    stop_trace, 
    start_scope, 
    stop_scope, 
    http_route_handler, 
    monocle_trace_scope, 
    amonocle_trace_scope,
    monocle_trace_scope_method, 
    monocle_trace, 
    amonocle_trace,
    monocle_trace_method,
    monocle_trace_http_route,
    is_valid_trace_id_uuid
)
from .utils import MonocleSpanException



from monocle_apptrace.instrumentation.common.wrapper import task_wrapper, atask_wrapper
from monocle_apptrace.instrumentation.metamodel.a2a.entities.inference import A2A_CLIENT

A2A_CLIENT_METHODS = [
    # {
    #     "package": "a2a.client.client",
    #     "object": "A2ACardResolver",
    #     "method": "get_agent_card",
    #     "wrapper_method": atask_wrapper,
    #     # "span_handler": "mcp_agent_handler",
    #     "output_processor": A2A_RESOLVE,
    # },
    {
        "package": "a2a.client.client",
        "object": "A2AClient",
        "method": "send_message",
        "wrapper_method": atask_wrapper,
        "output_processor": A2A_CLIENT,
    },
]





import logging
from threading import local
from monocle_apptrace.instrumentation.common.utils import extract_http_headers, clear_http_scopes
from monocle_apptrace.instrumentation.common.span_handler import SpanHandler
from monocle_apptrace.instrumentation.common.constants import HTTP_SUCCESS_CODES
from monocle_apptrace.instrumentation.common.utils import MonocleSpanException
from opentelemetry.context import get_current
from opentelemetry.trace import Span
from opentelemetry.trace.propagation import _SPAN_KEY
import json
import urllib.parse

logger = logging.getLogger(__name__)
MAX_DATA_LENGTH = 1000
token_data = local()
token_data.current_token = None

def get_route(scope) -> str:
    return scope.get('path', '')

def get_method(scope) -> str:
    return scope.get('method', '')

def get_params(args) -> dict:
    try:
        query_bytes = args.get("query_string", "")
        query_str = query_bytes.decode('utf-8')
        params = urllib.parse.parse_qs(query_str)
        question = params.get('question', [''])[0]
        return question
    except Exception as e:
        logger.warning(f"Error extracting params: {e}")
        return {}

def extract_response(response) -> str:
    try:
        if hasattr(response, 'body'):
            data = response.body
            answer = json.loads(data.decode("utf-8"))
            return answer
    except Exception as e:
        logger.warning(f"Error extracting response: {e}")
        return ""

def extract_status(instance) -> str:
    status = f"{instance.status_code}" if hasattr(instance, 'status_code') else ""
    if status not in HTTP_SUCCESS_CODES:
        error_message = extract_response(instance)
        raise MonocleSpanException(f"error: {status} - {error_message}")
    return status

def fastapi_pre_tracing(scope):
    headers = {k.decode('utf-8').lower(): v.decode('utf-8')
               for k, v in scope.get('headers', [])}
    token_data.current_token = extract_http_headers(headers)

def fastapi_post_tracing():
    clear_http_scopes(token_data.current_token)
    token_data.current_token = None

class FastAPISpanHandler(SpanHandler):
    def pre_tracing(self, to_wrap, wrapped, instance, args, kwargs):
        scope = args[0] if args else {}
        fastapi_pre_tracing(scope)
        return super().pre_tracing(to_wrap, wrapped, instance, args, kwargs)

    def post_tracing(self, to_wrap, wrapped, instance, args, kwargs, return_value):
        fastapi_post_tracing()
        return super().post_tracing(to_wrap, wrapped, instance, args, kwargs, return_value)

class FastAPIResponseSpanHandler(SpanHandler):
    def post_tracing(self, to_wrap, wrapped, instance, args, kwargs, return_value):
        try:
            ctx = get_current()
            if ctx is not None:
                parent_span: Span = ctx.get(_SPAN_KEY)
                if parent_span is not None:
                    self.hydrate_events(to_wrap, wrapped, instance, args, kwargs,
                                        return_value, parent_span=parent_span)
        except Exception as e:
            logger.info(f"Failed to propagate fastapi response: {e}")
        super().post_tracing(to_wrap, wrapped, instance, args, kwargs, return_value)


from monocle_apptrace.instrumentation.metamodel.gemini import (
    _helper,
)
from monocle_apptrace.instrumentation.common.utils import get_error_message

INFERENCE = {
    "type": "inference",
    "attributes": [
        [
            {
                "_comment": "provider type  , inference_endpoint",
                "attribute": "type",
                "accessor": lambda arguments: 'inference.vertexai' if hasattr(arguments['instance'],"vertexai") and arguments['instance'].vertexai else 'inference.gemini'
            },
            {
                "attribute": "inference_endpoint",
                "accessor": lambda arguments: _helper.extract_inference_endpoint(arguments['instance'])
            }
        ],
        [
            {
                "_comment": "LLM Model",
                "attribute": "name",
                "accessor": lambda arguments: _helper.resolve_from_alias(arguments['kwargs'],
                                                                         ['model'])
            },
            {
                "attribute": "type",
                "accessor": lambda arguments: 'model.llm.' + _helper.resolve_from_alias(arguments['kwargs'],
                                                                                        ['model'])
            }
        ]
    ],
    "events": [
        {
         "name": "data.input",
         "attributes": [

             {
                 "_comment": "this is instruction and user query to LLM",
                 "attribute": "input",
                 "accessor": lambda arguments: _helper.extract_messages(arguments['kwargs'])
             }
         ]
         },
        {
            "name": "data.output",
            "attributes": [
                {
                    "attribute": "error_code",
                    "accessor": lambda arguments: get_error_message(arguments)
                },
                {
                    "attribute": "response",
                    "accessor": lambda arguments: _helper.extract_assistant_message(arguments)
                }
            ]
        },
        {
            "name": "metadata",
            "attributes": [
                {
                    "_comment": "this is metadata usage from LLM",
                    "accessor": lambda arguments: _helper.update_span_from_llm_response(arguments['result'], arguments['instance'])
                },
                {
                    "_comment": "finish reason from Gemini response",
                    "attribute": "finish_reason",
                    "accessor": lambda arguments: _helper.extract_finish_reason(arguments)
                },
                {
                    "_comment": "finish type mapped from finish reason",
                    "attribute": "finish_type",
                    "accessor": lambda arguments: _helper.map_finish_reason_to_finish_type(
                        _helper.extract_finish_reason(arguments)
                    )
                }
            ]
        }


    ]
}



from monocle_apptrace.instrumentation.common.wrapper import task_wrapper, atask_wrapper
from monocle_apptrace.instrumentation.metamodel.langgraph.entities.inference import (
    AGENT, TOOLS
)
LANGGRAPH_METHODS = [
    {
        "package": "langgraph.graph.state",
         "object": "CompiledStateGraph",
         "method": "invoke",
         "wrapper_method": task_wrapper,
        "span_handler":"langgraph_agent_handler",
         "output_processor": AGENT
    },
    {
        "package": "langgraph.graph.state",
         "object": "CompiledStateGraph",
         "method": "ainvoke",
         "wrapper_method": atask_wrapper,
        "span_handler":"langgraph_agent_handler",
         "output_processor": AGENT
    },
    {
        "package": "langchain_core.tools.base",
         "object": "BaseTool",
         "method": "run",
         "wrapper_method": task_wrapper,
        "span_handler":"langgraph_tool_handler",
         "output_processor": TOOLS
    },
    {
        "package": "langchain_core.tools.base",
         "object": "BaseTool",
         "method": "arun",
         "wrapper_method": atask_wrapper,
        "span_handler":"langgraph_tool_handler",
         "output_processor": TOOLS
    },    
]


from monocle_apptrace.instrumentation.metamodel.lambdafunc import _helper
LAMBDA_HTTP_PROCESSOR = {
    "type": "http.process",
    "attributes": [
        [
            {
                "_comment": "request method, request URI",
                "attribute": "method",
                "accessor": lambda arguments: _helper.get_method(arguments['args'])
            },
            {
                "_comment": "request method, request URI",
                "attribute": "route",
                "accessor": lambda arguments: _helper.get_route(arguments['args'])
            },
            {
                "_comment": "request method, request URI",
                "attribute": "body",
                "accessor": lambda arguments: _helper.get_body(arguments['args'])
            },
        ]
    ],
    "events": [
        {
            "name": "data.input",
            "attributes": [
                {
                    "_comment": "route params",
                    "attribute": "params",
                    "accessor": lambda arguments: _helper.get_params(arguments['args'])
                }
            ]
        },
        {
            "name": "data.output",
            "attributes": [
                {
                    "_comment": "status from HTTP response",
                    "attribute": "status",
                    "accessor": lambda arguments: _helper.extract_status(arguments['result'])
                },
                {
                    "_comment": "this is result from LLM",
                    "attribute": "response",
                    "accessor": lambda arguments: _helper.extract_response(arguments['result'])
                }
            ]
        }

    ]
}



import logging
import random
import time
from types import SimpleNamespace
from monocle_apptrace.instrumentation.metamodel.openai import (
    _helper,
)
from monocle_apptrace.instrumentation.common.utils import (
    get_error_message,
    patch_instance_method,
    resolve_from_alias
)

logger = logging.getLogger(__name__)


def _process_stream_item(item, state):
    """Process a single stream item and update state."""
    try:
        if hasattr(item, "type") and isinstance(item.type, str) and item.type.startswith("response."):
            if state["waiting_for_first_token"]:
                state["waiting_for_first_token"] = False
                state["first_token_time"] = time.time_ns()
            if item.type == "response.output_text.delta":
                state["accumulated_response"] += item.delta
            if item.type == "response.completed":
                state["stream_closed_time"] = time.time_ns()
                if hasattr(item, "response") and hasattr(item.response, "usage"):
                    state["token_usage"] = item.response.usage
        elif (
            hasattr(item, "choices")
            and item.choices
            and item.choices[0].delta
            and item.choices[0].delta.content
        ):
            if hasattr(item.choices[0].delta, "role") and item.choices[0].delta.role:
                state["role"] = item.choices[0].delta.role
            if state["waiting_for_first_token"]:
                state["waiting_for_first_token"] = False
                state["first_token_time"] = time.time_ns()

            state["accumulated_response"] += item.choices[0].delta.content
        elif hasattr(item, "object") and item.object == "chat.completion.chunk" and item.usage:
            # Handle the case where the response is a chunk
            state["token_usage"] = item.usage
            state["stream_closed_time"] = time.time_ns()
            # Capture finish_reason from the chunk
            if (
                hasattr(item, "choices")
                and item.choices
                and len(item.choices) > 0
                and hasattr(item.choices[0], 'finish_reason')
                and item.choices[0].finish_reason
            ):
                finish_reason = item.choices[0].finish_reason
                state["finish_reason"] = finish_reason

    except Exception as e:
        logger.warning(
            "Warning: Error occurred while processing stream item: %s",
            str(e),
        )
    finally:
        state["accumulated_temp_list"].append(item)


def _create_span_result(state, stream_start_time):
    """Create the span result object."""
    return SimpleNamespace(
        type="stream",
        timestamps={
            "role": state["role"],
            "data.input": int(stream_start_time),
            "data.output": int(state["first_token_time"]),
            "metadata": int(state["stream_closed_time"] or time.time_ns()),
        },
        output_text=state["accumulated_response"],
        usage=state["token_usage"],
        finish_reason=state["finish_reason"]
    )


def process_stream(to_wrap, response, span_processor):
    stream_start_time = time.time_ns()
    
    # Shared state for both sync and async processing
    state = {
        "waiting_for_first_token": True,
        "first_token_time": stream_start_time,
        "stream_closed_time": None,
        "accumulated_response": "",
        "token_usage": None,
        "accumulated_temp_list": [],
        "finish_reason": None,
        "role": "assistant",
    }

    if to_wrap and hasattr(response, "__iter__"):
        original_iter = response.__iter__

        def new_iter(self):
            for item in original_iter():
                _process_stream_item(item, state)
                yield item

            if span_processor:
                ret_val = _create_span_result(state, stream_start_time)
                span_processor(ret_val)

        patch_instance_method(response, "__iter__", new_iter)
        
    if to_wrap and hasattr(response, "__aiter__"):
        original_iter = response.__aiter__

        async def new_aiter(self):
            async for item in original_iter():
                _process_stream_item(item, state)
                yield item

            if span_processor:
                ret_val = _create_span_result(state, stream_start_time)
                span_processor(ret_val)

        patch_instance_method(response, "__aiter__", new_aiter)


INFERENCE = {
    "type": "inference",
    "is_auto_close": lambda kwargs: kwargs.get("stream", False) is False,
    "response_processor": process_stream,
    "attributes": [
        [
            {
                "_comment": "provider type ,name , deployment , inference_endpoint",
                "attribute": "type",
                "accessor": lambda arguments: "inference."
                + (_helper.get_inference_type(arguments["instance"]))
                or "openai",
            },
            {
                "attribute": "provider_name",
                "accessor": lambda arguments: _helper.extract_provider_name(
                    arguments["instance"]
                ),
            },
            {
                "attribute": "deployment",
                "accessor": lambda arguments: resolve_from_alias(
                    arguments["instance"].__dict__,
                    [
                        "engine",
                        "azure_deployment",
                        "deployment_name",
                        "deployment_id",
                        "deployment",
                    ],
                ),
            },
            {
                "attribute": "inference_endpoint",
                "accessor": lambda arguments: resolve_from_alias(
                    arguments["instance"].__dict__,
                    ["azure_endpoint", "api_base", "endpoint"],
                )
                or _helper.extract_inference_endpoint(arguments["instance"]),
            },
        ],
        [
            {
                "_comment": "LLM Model",
                "attribute": "name",
                "accessor": lambda arguments: resolve_from_alias(
                    arguments["kwargs"],
                    ["model", "model_name", "endpoint_name", "deployment_name"],
                ),
            },
            {
                "attribute": "type",
                "accessor": lambda arguments: "model.llm."
                + resolve_from_alias(
                    arguments["kwargs"],
                    ["model", "model_name", "endpoint_name", "deployment_name"],
                ),
            },
        ],
    ],
    "events": [
        {
            "name": "data.input",
            "attributes": [
                {
                    "_comment": "this is instruction and user query to LLM",
                    "attribute": "input",
                    "accessor": lambda arguments: _helper.extract_messages(
                        arguments["kwargs"]
                    ),
                }
            ],
        },
        {
            "name": "data.output",
            "attributes": [

                {
                    "attribute": "error_code",
                    "accessor": lambda arguments: get_error_message(arguments)
                },
                {
                    "_comment": "this is result from LLM",
                    "attribute": "response",
                    "accessor": lambda arguments: _helper.extract_assistant_message(
                        arguments,
                    ),
                }
            ],
        },
        {
            "name": "metadata",
            "attributes": [
                {
                    "_comment": "this is metadata usage from LLM",
                    "accessor": lambda arguments: _helper.update_span_from_llm_response(
                        arguments["result"]
                    ),
                },
                {
                    "_comment": "finish reason from OpenAI response",
                    "attribute": "finish_reason",
                    "accessor": lambda arguments: _helper.extract_finish_reason(arguments)
                },
                {
                    "_comment": "finish type mapped from finish reason",
                    "attribute": "finish_type",
                    "accessor": lambda arguments: _helper.map_finish_reason_to_finish_type(
                        _helper.extract_finish_reason(arguments)
                    )
                }
            ],
        },
    ],
}



import logging
from threading import local
from monocle_apptrace.instrumentation.common.utils import extract_http_headers, clear_http_scopes, try_option, Option, MonocleSpanException
from monocle_apptrace.instrumentation.common.span_handler import SpanHandler
from monocle_apptrace.instrumentation.common.constants import HTTP_SUCCESS_CODES
from urllib.parse import unquote

logger = logging.getLogger(__name__)
MAX_DATA_LENGTH = 1000

def get_route(args) -> str:
    route_path: Option[str] = try_option(getattr, args[0], 'path')
    return route_path.unwrap_or("")

def get_method(args) -> str:
#    return args[0]['method'] if 'method' in args[0] else ""
    http_method: Option[str] = try_option(getattr, args[0], 'method')
    return http_method.unwrap_or("")

def get_params(args) -> dict:
    params: Option[str] = try_option(getattr, args[0], 'query_string')
    return unquote(params.unwrap_or(""))

def get_body(args) -> dict:
    return ""

def extract_response(result) -> str:
    if hasattr(result, 'text'):
        response = result.text[0:max(result.text.__len__(), MAX_DATA_LENGTH)]
    else:
        response = ""
    return response

def extract_status(result) -> str:
    status = f"{result.status}" if hasattr(result, 'status') else ""
    if status not in HTTP_SUCCESS_CODES:
        error_message = extract_response(result)
        raise MonocleSpanException(f"error: {status} - {error_message}")
    return status

def aiohttp_pre_tracing(args):
    return extract_http_headers(args[0].headers)

def aiohttp_post_tracing(token):
    clear_http_scopes(token)

def aiohttp_skip_span(args) -> bool:
    if get_method(args) == "HEAD":
        return True
    return False

class aiohttpSpanHandler(SpanHandler):

    def pre_tracing(self, to_wrap, wrapped, instance, args, kwargs):
        return aiohttp_pre_tracing(args)
    
    def post_tracing(self, to_wrap, wrapped, instance, args, kwargs, return_value, token):
        aiohttp_post_tracing(token)

    def skip_span(self, to_wrap, wrapped, instance, args, kwargs) -> bool:
        return aiohttp_skip_span(args)


from monocle_apptrace.instrumentation.common.wrapper import atask_wrapper
from monocle_apptrace.instrumentation.metamodel.fastapi.entities.http import FASTAPI_HTTP_PROCESSOR, FASTAPI_RESPONSE_PROCESSOR

FASTAPI_METHODS = [
    {
        "package": "fastapi",
        "object": "FastAPI",
        "method": "__call__",
        "wrapper_method": atask_wrapper,
        "span_name": "fastapi.request",
        "span_handler": "fastapi_handler",
        "output_processor": FASTAPI_HTTP_PROCESSOR,
    },
    {
        "package": "starlette.responses",
        "object": "Response",
        "method": "__call__",
        "span_name": "fastapi.response",
        "wrapper_method": atask_wrapper,
        "span_handler": "fastapi_response_handler",
        "output_processor": FASTAPI_RESPONSE_PROCESSOR
    }
]


"""
This module provides common finish reason mappings and finish type enums
for different AI providers (OpenAI, Anthropic, Gemini, LangChain, LlamaIndex, Azure AI Inference).
"""

from enum import Enum

class FinishType(Enum):
    """Enum for standardized finish types across all AI providers."""
    SUCCESS = "success"
    TRUNCATED = "truncated"
    CONTENT_FILTER = "content_filter"
    ERROR = "error"
    REFUSAL = "refusal"
    RATE_LIMITED = "rate_limited"

# OpenAI finish reason mapping
OPENAI_FINISH_REASON_MAPPING = {
    "stop": FinishType.SUCCESS.value,
    "tool_calls": FinishType.SUCCESS.value,
    "function_call": FinishType.SUCCESS.value,  # deprecated but still possible
    "length": FinishType.TRUNCATED.value,
    "content_filter": FinishType.CONTENT_FILTER.value
}

# Anthropic finish reason mapping
ANTHROPIC_FINISH_REASON_MAPPING = {
    "end_turn": FinishType.SUCCESS.value,         # Natural completion
    "max_tokens": FinishType.TRUNCATED.value,     # Hit max_tokens limit
    "stop_sequence": FinishType.SUCCESS.value,    # Hit user stop sequence
    "tool_use": FinishType.SUCCESS.value,         # Tool use triggered
    "pause_turn": FinishType.SUCCESS.value,       # Paused for tool or server action
    "refusal": FinishType.REFUSAL.value,          # Refused for safety/ethics
}

# Gemini finish reason mapping
GEMINI_FINISH_REASON_MAPPING = {
    "STOP": FinishType.SUCCESS.value,
    "MAX_TOKENS": FinishType.TRUNCATED.value,
    "SAFETY": FinishType.CONTENT_FILTER.value,
    "RECITATION": FinishType.CONTENT_FILTER.value,
    "OTHER": FinishType.ERROR.value,
    "FINISH_REASON_UNSPECIFIED": None
}

# LlamaIndex finish reason mapping
# LlamaIndex often wraps underlying provider responses, similar to LangChain
LLAMAINDEX_FINISH_REASON_MAPPING = {
    # Standard completion reasons
    "stop": FinishType.SUCCESS.value,
    "complete": FinishType.SUCCESS.value,
    "finished": FinishType.SUCCESS.value,
    "success": FinishType.SUCCESS.value,
    
    # Token limits
    "length": FinishType.TRUNCATED.value,
    "max_tokens": FinishType.TRUNCATED.value,
    "token_limit": FinishType.TRUNCATED.value,
    "truncated": FinishType.TRUNCATED.value,
    
    # Tool/function calling
    "tool_calls": FinishType.SUCCESS.value,
    "function_call": FinishType.SUCCESS.value,
    "agent_finish": FinishType.SUCCESS.value,
    
    # Content filtering and safety
    "content_filter": FinishType.CONTENT_FILTER.value,
    "safety": FinishType.CONTENT_FILTER.value,
    "filtered": FinishType.CONTENT_FILTER.value,
    
    # Errors
    "error": FinishType.ERROR.value,
    "failed": FinishType.ERROR.value,
    "exception": FinishType.ERROR.value,
    
    # Provider-specific reasons that might pass through LlamaIndex
    # OpenAI reasons
    "end_turn": FinishType.SUCCESS.value,  # Anthropic
    "stop_sequence": FinishType.SUCCESS.value,  # Anthropic
    "STOP": FinishType.SUCCESS.value,  # Gemini
    "SAFETY": FinishType.CONTENT_FILTER.value,  # Gemini
    "RECITATION": FinishType.CONTENT_FILTER.value,  # Gemini
    "OTHER": FinishType.ERROR.value,  # Gemini
}

# Azure AI Inference finish reason mapping
AZURE_AI_INFERENCE_FINISH_REASON_MAPPING = {
    # Standard completion reasons
    "stop": FinishType.SUCCESS.value,
    "completed": FinishType.SUCCESS.value,
    "finished": FinishType.SUCCESS.value,
    
    # Token limits
    "length": FinishType.TRUNCATED.value,
    "max_tokens": FinishType.TRUNCATED.value,
    "token_limit": FinishType.TRUNCATED.value,
    "max_completion_tokens": FinishType.TRUNCATED.value,
    
    # Tool/function calling
    "tool_calls": FinishType.SUCCESS.value,
    "function_call": FinishType.SUCCESS.value,
    
    # Content filtering and safety
    "content_filter": FinishType.CONTENT_FILTER.value,
    "content_filtered": FinishType.CONTENT_FILTER.value,
    "safety": FinishType.CONTENT_FILTER.value,
    "responsible_ai_policy": FinishType.CONTENT_FILTER.value,
    
    # Errors
    "error": FinishType.ERROR.value,
    "failed": FinishType.ERROR.value,
    "exception": FinishType.ERROR.value,
    "timeout": FinishType.ERROR.value,
    
    # Azure-specific reasons
    "model_error": FinishType.ERROR.value,
    "service_unavailable": FinishType.ERROR.value,
    "rate_limit": FinishType.ERROR.value,
}

# AWS Bedrock finish reason mapping
# Based on AWS Bedrock Converse API and model-specific APIs
BEDROCK_FINISH_REASON_MAPPING = {
    # Standard completion reasons
    "end_turn": FinishType.SUCCESS.value,           # Natural completion
    "stop": FinishType.SUCCESS.value,               # Hit stop sequence
    "stop_sequence": FinishType.SUCCESS.value,      # Stop sequence triggered
    "completed": FinishType.SUCCESS.value,          # Completion finished successfully
    
    # Token limits
    "max_tokens": FinishType.TRUNCATED.value,       # Hit max_tokens limit
    "length": FinishType.TRUNCATED.value,           # Token length limit
    "max_length": FinishType.TRUNCATED.value,       # Maximum length reached
    "token_limit": FinishType.TRUNCATED.value,      # Token limit reached
    
    # Tool/function calling
    "tool_use": FinishType.SUCCESS.value,           # Tool use triggered
    "function_call": FinishType.SUCCESS.value,      # Function call triggered
    
    # Content filtering and safety
    "content_filter": FinishType.CONTENT_FILTER.value,    # Content filtered
    "content_filtered": FinishType.CONTENT_FILTER.value,  # Content was filtered
    "safety": FinishType.CONTENT_FILTER.value,            # Safety filter triggered
    "guardrails": FinishType.CONTENT_FILTER.value,        # Bedrock guardrails triggered
    "blocked": FinishType.CONTENT_FILTER.value,           # Request blocked
    
    # Errors
    "error": FinishType.ERROR.value,                # General error
    "failed": FinishType.ERROR.value,               # Request failed
    "exception": FinishType.ERROR.value,            # Exception occurred
    "timeout": FinishType.ERROR.value,              # Request timeout
    "model_error": FinishType.ERROR.value,          # Model-specific error
    "service_unavailable": FinishType.ERROR.value,  # Service unavailable
    "throttled": FinishType.ERROR.value,            # Request throttled
    "rate_limit": FinishType.ERROR.value,           # Rate limit exceeded
    "validation_error": FinishType.ERROR.value,     # Validation error
    
    # Model-specific reasons (various Bedrock models)
    # Claude models via Bedrock
    "end_turn": FinishType.SUCCESS.value,           # Already defined above
    "max_tokens": FinishType.TRUNCATED.value,       # Already defined above
    "stop_sequence": FinishType.SUCCESS.value,      # Already defined above
    "tool_use": FinishType.SUCCESS.value,           # Already defined above
    
    # AI21 models via Bedrock
    "endoftext": FinishType.SUCCESS.value,          # AI21 end of text
    "length": FinishType.TRUNCATED.value,           # AI21 length limit
    
    # Cohere models via Bedrock
    "COMPLETE": FinishType.SUCCESS.value,           # Cohere completion
    "MAX_TOKENS": FinishType.TRUNCATED.value,       # Cohere max tokens
    "ERROR": FinishType.ERROR.value,                # Cohere error
    
    # Meta Llama models via Bedrock
    "stop": FinishType.SUCCESS.value,               # Already defined above
    "length": FinishType.TRUNCATED.value,           # Already defined above
    
    # Amazon Titan models via Bedrock
    "FINISH": FinishType.SUCCESS.value,             # Titan finish
    "LENGTH": FinishType.TRUNCATED.value,           # Titan length limit
    "CONTENT_FILTERED": FinishType.CONTENT_FILTER.value,  # Titan content filter
}

# LangChain finish reason mapping
# LangChain often wraps underlying provider responses, so we include common finish reasons
# that might appear in LangChain response objects
LANGCHAIN_FINISH_REASON_MAPPING = {
    # Standard completion reasons
    "stop": FinishType.SUCCESS.value,
    "complete": FinishType.SUCCESS.value,
    "finished": FinishType.SUCCESS.value,
    
    # Token limits
    "length": FinishType.TRUNCATED.value,
    "max_tokens": FinishType.TRUNCATED.value,
    "token_limit": FinishType.TRUNCATED.value,
    
    # Tool/function calling
    "tool_calls": FinishType.SUCCESS.value,
    "function_call": FinishType.SUCCESS.value,
    
    # Content filtering and safety
    "content_filter": FinishType.CONTENT_FILTER.value,
    "safety": FinishType.CONTENT_FILTER.value,
    "filtered": FinishType.CONTENT_FILTER.value,
    
    # Errors
    "error": FinishType.ERROR.value,
    "failed": FinishType.ERROR.value,
    "exception": FinishType.ERROR.value,
    
    # Provider-specific reasons that might pass through LangChain
    # OpenAI reasons
    "stop": FinishType.SUCCESS.value,  # Already defined above
    
    # Anthropic reasons
    "end_turn": FinishType.SUCCESS.value,
    "stop_sequence": FinishType.SUCCESS.value,
    
    # Gemini reasons
    "STOP": FinishType.SUCCESS.value,
    "SAFETY": FinishType.CONTENT_FILTER.value,
    "RECITATION": FinishType.CONTENT_FILTER.value,
    "OTHER": FinishType.ERROR.value,
}

TEAMSAI_FINISH_REASON_MAPPING = {
    "success": FinishType.SUCCESS.value,
    "error": FinishType.ERROR.value,
    "too_long": FinishType.TRUNCATED.value,
    "rate_limited": FinishType.RATE_LIMITED.value,
    "invalid_response": FinishType.ERROR.value,
}

def map_openai_finish_reason_to_finish_type(finish_reason):
    """Map OpenAI finish_reason to standardized finish_type."""
    if not finish_reason:
        return None
    return OPENAI_FINISH_REASON_MAPPING.get(finish_reason, None)


def map_anthropic_finish_reason_to_finish_type(finish_reason):
    """Map Anthropic stop_reason to standardized finish_type."""
    if not finish_reason:
        return None
    return ANTHROPIC_FINISH_REASON_MAPPING.get(finish_reason, None)


def map_gemini_finish_reason_to_finish_type(finish_reason):
    """Map Gemini finish_reason to standardized finish_type."""
    if not finish_reason:
        return None
    return GEMINI_FINISH_REASON_MAPPING.get(finish_reason, None)


def map_langchain_finish_reason_to_finish_type(finish_reason):
    """Map LangChain finish_reason to standardized finish_type."""
    if not finish_reason:
        return None
    
    # Convert to lowercase for case-insensitive matching
    finish_reason_lower = finish_reason.lower() if isinstance(finish_reason, str) else str(finish_reason).lower()
    
    # Try direct mapping first
    if finish_reason in LANGCHAIN_FINISH_REASON_MAPPING:
        return LANGCHAIN_FINISH_REASON_MAPPING[finish_reason]
    
    # Try lowercase mapping
    if finish_reason_lower in LANGCHAIN_FINISH_REASON_MAPPING:
        return LANGCHAIN_FINISH_REASON_MAPPING[finish_reason_lower]
    
    # If no direct mapping, try to infer from common patterns
    if any(keyword in finish_reason_lower for keyword in ['stop', 'complete', 'success', 'done']):
        return FinishType.SUCCESS.value
    elif any(keyword in finish_reason_lower for keyword in ['length', 'token', 'limit', 'truncat']):
        return FinishType.TRUNCATED.value
    elif any(keyword in finish_reason_lower for keyword in ['filter', 'safety', 'block']):
        return FinishType.CONTENT_FILTER.value
    elif any(keyword in finish_reason_lower for keyword in ['error', 'fail', 'exception']):
        return FinishType.ERROR.value
    
    return None


def map_llamaindex_finish_reason_to_finish_type(finish_reason):
    """Map LlamaIndex finish_reason to standardized finish_type."""
    if not finish_reason:
        return None
    
    # Convert to lowercase for case-insensitive matching
    finish_reason_lower = finish_reason.lower() if isinstance(finish_reason, str) else str(finish_reason).lower()
    
    # Try direct mapping first
    if finish_reason in LLAMAINDEX_FINISH_REASON_MAPPING:
        return LLAMAINDEX_FINISH_REASON_MAPPING[finish_reason]
    
    # Try lowercase mapping
    if finish_reason_lower in LLAMAINDEX_FINISH_REASON_MAPPING:
        return LLAMAINDEX_FINISH_REASON_MAPPING[finish_reason_lower]
    
    # If no direct mapping, try to infer from common patterns
    if any(keyword in finish_reason_lower for keyword in ['stop', 'complete', 'success', 'done', 'finish']):
        return FinishType.SUCCESS.value
    elif any(keyword in finish_reason_lower for keyword in ['length', 'token', 'limit', 'truncat']):
        return FinishType.TRUNCATED.value
    elif any(keyword in finish_reason_lower for keyword in ['filter', 'safety', 'block']):
        return FinishType.CONTENT_FILTER.value
    elif any(keyword in finish_reason_lower for keyword in ['error', 'fail', 'exception']):
        return FinishType.ERROR.value
    
    return None


def map_azure_ai_inference_finish_reason_to_finish_type(finish_reason):
    """Map Azure AI Inference finish_reason to standardized finish_type."""
    if not finish_reason:
        return None
    
    # Convert to lowercase for case-insensitive matching
    finish_reason_lower = finish_reason.lower() if isinstance(finish_reason, str) else str(finish_reason).lower()
    
    # Try direct mapping first
    if finish_reason in AZURE_AI_INFERENCE_FINISH_REASON_MAPPING:
        return AZURE_AI_INFERENCE_FINISH_REASON_MAPPING[finish_reason]
    
    # Try lowercase mapping
    if finish_reason_lower in AZURE_AI_INFERENCE_FINISH_REASON_MAPPING:
        return AZURE_AI_INFERENCE_FINISH_REASON_MAPPING[finish_reason_lower]
    
    # If no direct mapping, try to infer from common patterns
    if any(keyword in finish_reason_lower for keyword in ['stop', 'complete', 'success', 'done', 'finish']):
        return FinishType.SUCCESS.value
    elif any(keyword in finish_reason_lower for keyword in ['length', 'token', 'limit', 'truncat']):
        return FinishType.TRUNCATED.value
    elif any(keyword in finish_reason_lower for keyword in ['filter', 'safety', 'block', 'responsible_ai', 'content_filter']):
        return FinishType.CONTENT_FILTER.value
    elif any(keyword in finish_reason_lower for keyword in ['error', 'fail', 'exception', 'timeout', 'unavailable', 'rate_limit']):
        return FinishType.ERROR.value
    
    return None


def map_bedrock_finish_reason_to_finish_type(finish_reason):
    """Map AWS Bedrock finish_reason/stopReason to standardized finish_type."""
    if not finish_reason:
        return None
    
    # Convert to lowercase for case-insensitive matching
    finish_reason_lower = finish_reason.lower() if isinstance(finish_reason, str) else str(finish_reason).lower()
    
    # Try direct mapping first
    if finish_reason in BEDROCK_FINISH_REASON_MAPPING:
        return BEDROCK_FINISH_REASON_MAPPING[finish_reason]
    
    # Try lowercase mapping
    if finish_reason_lower in BEDROCK_FINISH_REASON_MAPPING:
        return BEDROCK_FINISH_REASON_MAPPING[finish_reason_lower]
    
    # If no direct mapping, try to infer from common patterns
    if any(keyword in finish_reason_lower for keyword in ['stop', 'complete', 'success', 'done', 'finish', 'end_turn', 'endoftext']):
        return FinishType.SUCCESS.value
    elif any(keyword in finish_reason_lower for keyword in ['length', 'token', 'limit', 'truncat', 'max_tokens']):
        return FinishType.TRUNCATED.value
    elif any(keyword in finish_reason_lower for keyword in ['filter', 'safety', 'block', 'guardrails', 'content_filter']):
        return FinishType.CONTENT_FILTER.value
    elif any(keyword in finish_reason_lower for keyword in ['error', 'fail', 'exception', 'timeout', 'unavailable', 'rate_limit', 'throttled', 'validation']):
        return FinishType.ERROR.value
    
    return None

def map_teamsai_finish_reason_to_finish_type(finish_reason):
    """Map TeamsAI finish_reason to standardized finish_type."""
    if not finish_reason:
        return None

    # Convert to lowercase for case-insensitive matching
    finish_reason_lower = finish_reason.lower() if isinstance(finish_reason, str) else str(finish_reason).lower()

    # Try direct mapping first
    if finish_reason in TEAMSAI_FINISH_REASON_MAPPING:
        return TEAMSAI_FINISH_REASON_MAPPING[finish_reason]

    # Try lowercase mapping
    if finish_reason_lower in TEAMSAI_FINISH_REASON_MAPPING:
        return TEAMSAI_FINISH_REASON_MAPPING[finish_reason_lower]

    return None


import logging
import os
from contextlib import contextmanager
from opentelemetry.context import get_value, set_value, attach, detach
from opentelemetry.sdk.trace import Span
from opentelemetry.trace.status import Status, StatusCode
from monocle_apptrace.instrumentation.common.constants import (
    QUERY,
    service_name_map,
    service_type_map,
    MONOCLE_SDK_VERSION, MONOCLE_SDK_LANGUAGE, MONOCLE_DETECTED_SPAN_ERROR
)
from monocle_apptrace.instrumentation.common.utils import set_attribute, get_scopes, MonocleSpanException, get_monocle_version
from monocle_apptrace.instrumentation.common.constants import WORKFLOW_TYPE_KEY, WORKFLOW_TYPE_GENERIC, CHILD_ERROR_CODE

logger = logging.getLogger(__name__)

WORKFLOW_TYPE_MAP = {
    "llama_index.core.agent.workflow": WORKFLOW_TYPE_GENERIC,
    "llama_index": "workflow.llamaindex",
    "langchain": "workflow.langchain",
    "haystack": "workflow.haystack",
    "teams.ai": "workflow.teams_ai",
    "langgraph": "workflow.langgraph",
    "openai": "workflow.openai",
    "anthropic": "workflow.anthropic",
    "gemini": "workflow.gemini",
}

FRAMEWORK_WORKFLOW_LIST = [
    "workflow.llamaindex",
    "workflow.langchain",
    "workflow.haystack",
    "workflow.teams_ai",
]
class SpanHandler:

    def __init__(self,instrumentor=None):
        self.instrumentor=instrumentor

    def set_instrumentor(self, instrumentor):
        self.instrumentor = instrumentor

    def validate(self, to_wrap, wrapped, instance, args, kwargs):
        pass

    def pre_tracing(self, to_wrap, wrapped, instance, args, kwargs):
        pass

    def post_tracing(self, to_wrap, wrapped, instance, args, kwargs, return_value, token=None):
        pass

    def skip_span(self, to_wrap, wrapped, instance, args, kwargs) -> bool:
        return False

    def skip_processor(self, to_wrap, wrapped, instance, span, args, kwargs) -> list[str]:
        return []
    
    def set_span_type(self, to_wrap, wrapped, instance, output_processor, span:Span, args, kwargs) -> str:
        span_type:str = None
        if 'type' in output_processor:
            span_type = output_processor['type']
            span.set_attribute("span.type", span_type)
        else:
            logger.warning("type of span not found or incorrect written in entity json")
        return span_type

    def pre_task_processing(self, to_wrap, wrapped, instance, args,kwargs, span):
        if "pipeline" in to_wrap['package']:
            set_attribute(QUERY, args[0]['prompt_builder']['question'])

    @staticmethod
    def set_default_monocle_attributes(span: Span, source_path = "" ):
        """ Set default monocle attributes for all spans """
        span.set_attribute(MONOCLE_SDK_VERSION, get_monocle_version())
        span.set_attribute(MONOCLE_SDK_LANGUAGE, "python")
        span.set_attribute("span_source", source_path)
        for scope_key, scope_value in get_scopes().items():
            span.set_attribute(f"scope.{scope_key}", scope_value)
        workflow_name = SpanHandler.get_workflow_name(span=span)
        if workflow_name:
            span.set_attribute("workflow.name", workflow_name)

    @staticmethod
    def set_workflow_properties(span: Span, to_wrap = None):
        """ Set attributes of workflow if this is a root span"""
        SpanHandler.set_workflow_attributes(to_wrap, span)
        SpanHandler.set_app_hosting_identifier_attribute(span)

    @staticmethod
    def set_non_workflow_properties(span: Span, to_wrap = None):
        span.set_attribute("span.type", "generic")

    def post_task_processing(self, to_wrap, wrapped, instance, args, kwargs, result, ex, span:Span, parent_span:Span):
        pass

    def hydrate_span(self, to_wrap, wrapped, instance, args, kwargs, result, span, parent_span = None, ex:Exception = None) -> bool:
        try:
            detected_error_in_attribute = self.hydrate_attributes(to_wrap, wrapped, instance, args, kwargs, result, span, parent_span)
            detected_error_in_event = self.hydrate_events(to_wrap, wrapped, instance, args, kwargs, result, span, parent_span, ex)
            if detected_error_in_attribute or detected_error_in_event:
                span.set_attribute(MONOCLE_DETECTED_SPAN_ERROR, True)
        finally:
            if span.status.status_code == StatusCode.UNSET and ex is None:
                span.set_status(StatusCode.OK)

    def hydrate_attributes(self, to_wrap, wrapped, instance, args, kwargs, result, span:Span, parent_span:Span) -> bool:
        detected_error:bool = False
        span_index = 0
        if SpanHandler.is_root_span(span):
            span_index = 2 # root span will have workflow and hosting entities pre-populated
        if 'output_processor' in to_wrap and to_wrap["output_processor"] is not None:
            output_processor=to_wrap['output_processor']
            self.set_span_type(to_wrap, wrapped, instance, output_processor, span, args, kwargs)
            skip_processors:list[str] = self.skip_processor(to_wrap, wrapped, instance, span, args, kwargs) or []

            if 'attributes' in output_processor and 'attributes' not in skip_processors:
                arguments = {"instance":instance, "args":args, "kwargs":kwargs, "result":result, "parent_span":parent_span, "span":span}
                for processors in output_processor["attributes"]:
                    for processor in processors:
                        attribute = processor.get('attribute')
                        accessor = processor.get('accessor')

                        if attribute and accessor:
                            attribute_name = f"entity.{span_index+1}.{attribute}"
                            try:
                                processor_result = accessor(arguments)
                                if processor_result and isinstance(processor_result, (str, list)):
                                    span.set_attribute(attribute_name, processor_result)
                            except MonocleSpanException as e:
                                span.set_status(StatusCode.ERROR, e.message)
                                detected_error = True
                            except Exception as e:
                                logger.debug(f"Error processing accessor: {e}")
                        else:
                            logger.debug(f"{' and '.join([key for key in ['attribute', 'accessor'] if not processor.get(key)])} not found or incorrect in entity JSON")
                    span_index += 1

        # set scopes as attributes by calling get_scopes()
        # scopes is a Mapping[str:object], iterate directly with .items()
        for scope_key, scope_value in get_scopes().items():
            span.set_attribute(f"scope.{scope_key}", scope_value)
        
        if span_index > 0:
            span.set_attribute("entity.count", span_index)
        return detected_error

    def hydrate_events(self, to_wrap, wrapped, instance, args, kwargs, ret_result, span, parent_span=None, ex:Exception=None) -> bool:
        detected_error:bool = False
        if 'output_processor' in to_wrap and to_wrap["output_processor"] is not None:
            output_processor=to_wrap['output_processor']
            skip_processors:list[str] = self.skip_processor(to_wrap, wrapped, instance, span, args, kwargs) or []

            arguments = {"instance": instance, "args": args, "kwargs": kwargs, "result": ret_result, "exception":ex, "parent_span":parent_span, "span": span}
            # Process events if they are defined in the output_processor.
            # In case of inference.modelapi skip the event processing unless the span has an exception
            if 'events' in output_processor and ('events' not in skip_processors or ex is not None):
                events = output_processor['events']
                for event in events:
                    event_name = event.get("name")
                    if 'events.'+event_name in skip_processors and ex is None:
                        continue
                    event_attributes = {}
                    attributes = event.get("attributes", [])
                    for attribute in attributes:
                        attribute_key = attribute.get("attribute")
                        accessor = attribute.get("accessor")
                        if accessor:
                            try:
                                result = accessor(arguments)
                                if result and isinstance(result, dict):
                                    result = dict((key, value) for key, value in result.items() if value is not None)
                                if result and isinstance(result, (int, str, list, dict)):
                                    if attribute_key is not None:
                                        event_attributes[attribute_key] = result
                                    else:
                                        event_attributes.update(result)
                            except MonocleSpanException as e:
                                span.set_status(StatusCode.ERROR, e.message)
                                detected_error = True
                            except Exception as e:
                                logger.debug(f"Error evaluating accessor for attribute '{attribute_key}': {e}")
                    matching_timestamp = getattr(ret_result, "timestamps", {}).get(event_name, None)
                    if isinstance(matching_timestamp, int):
                        span.add_event(name=event_name, attributes=event_attributes, timestamp=matching_timestamp)
                    else:
                        span.add_event(name=event_name, attributes=event_attributes)
        return detected_error

    @staticmethod
    def set_workflow_attributes(to_wrap, span: Span):
        span_index = 1
        workflow_name = SpanHandler.get_workflow_name(span=span)
        if workflow_name:
            span.update_name("workflow")
            span.set_attribute("span.type", "workflow")
            span.set_attribute(f"entity.{span_index}.name", workflow_name)
        workflow_type = SpanHandler.get_workflow_type(to_wrap)
        span.set_attribute(f"entity.{span_index}.type", workflow_type)

    def get_workflow_name_in_progress(self) -> str:
        return get_value(WORKFLOW_TYPE_KEY)

    @staticmethod
    def is_framework_workflow(workflow_type) -> bool:
        return workflow_type in FRAMEWORK_WORKFLOW_LIST

    def is_framework_span_in_progress(self) -> bool:
        return SpanHandler.is_framework_workflow(self.get_workflow_name_in_progress())

    @staticmethod
    def get_workflow_type(to_wrap):
        # workflow type
        workflow_type = WORKFLOW_TYPE_GENERIC
        if to_wrap is not None:
            package_name = to_wrap.get('package')
            for (package, framework_workflow_type) in WORKFLOW_TYPE_MAP.items():
                if (package_name is not None and package in package_name):
                    workflow_type = framework_workflow_type
                    break
        return workflow_type

    def set_app_hosting_identifier_attribute(span):
        span_index = 2
        # Search env to indentify the infra service type, if found check env for service name if possible
        span.set_attribute(f"entity.{span_index}.type", f"app_hosting.generic")
        span.set_attribute(f"entity.{span_index}.name", "generic")
        for type_env, type_name in service_type_map.items():
            if type_env in os.environ:
                span.set_attribute(f"entity.{span_index}.type", f"app_hosting.{type_name}")
                entity_name_env = service_name_map.get(type_name, "unknown")
                span.set_attribute(f"entity.{span_index}.name", os.environ.get(entity_name_env, "generic"))

    @staticmethod
    def get_workflow_name(span: Span) -> str:
        try:
            return get_value("workflow_name") or span.resource.attributes.get("service.name")
        except Exception as e:
            logger.exception(f"Error getting workflow name: {e}")
            return None

    @staticmethod
    def is_root_span(curr_span: Span) -> bool:
        try:
            if curr_span is not None and hasattr(curr_span, "parent") or  curr_span.context.trace_state:
                return curr_span.parent is None
        except Exception as e:
            logger.warning(f"Error finding root span: {e}")

    @staticmethod
    def attach_workflow_type(to_wrap=None, context=None): 
        token = None
        if to_wrap:
            workflow_type = SpanHandler.get_workflow_type(to_wrap)
            if SpanHandler.is_framework_workflow(workflow_type):
                token = attach(set_value(WORKFLOW_TYPE_KEY,
                                        SpanHandler.get_workflow_type(to_wrap), context))
        return token

    @staticmethod
    def detach_workflow_type(token):
        if token:
            return detach(token)

    @staticmethod
    @contextmanager
    def workflow_type(to_wrap=None, span:Span=None):
        token = SpanHandler.attach_workflow_type(to_wrap)
        try:
            yield
        finally:
            SpanHandler.detach_workflow_type(token)


class NonFrameworkSpanHandler(SpanHandler):
    # If the language framework is being executed, then skip generating direct openAI attributes and events
    def skip_processor(self, to_wrap, wrapped, instance, span, args, kwargs) -> list[str]:
        if super().is_framework_span_in_progress():
            return ["attributes", "events"]
    
    def set_span_type(self, to_wrap, wrapped, instance, output_processor, span:Span, args, kwargs) -> str:
        span_type = super().set_span_type(to_wrap, wrapped, instance, output_processor, span, args, kwargs)
        if self.is_framework_span_in_progress() and span_type is not None:
            span_type = span_type+".modelapi"
            span.set_attribute("span.type", span_type)
        return span_type




from opentelemetry.context import set_value, attach, detach
from monocle_apptrace.instrumentation.common.span_handler import SpanHandler
from monocle_apptrace.instrumentation.metamodel.langgraph._helper import (
   get_name, is_root_agent_name, is_delegation_tool, LANGGRAPTH_AGENT_NAME_KEY

)
from monocle_apptrace.instrumentation.metamodel.langgraph.entities.inference import (
    AGENT_DELEGATION, AGENT_REQUEST
)

class LanggraphAgentHandler(SpanHandler):
    def pre_tracing(self, to_wrap, wrapped, instance, args, kwargs):
        return attach(set_value(LANGGRAPTH_AGENT_NAME_KEY, get_name(instance)))

    def post_tracing(self, to_wrap, wrapped, instance, args, kwargs, result, token):
        if token is not None:
            detach(token)

    # In multi agent scenarios, the root agent is the one that orchestrates the other agents. LangGraph generates an extra root level invoke()
    # call on top of the supervisor agent invoke().
    # This span handler resets the parent invoke call as generic type to avoid duplicate attributes/events in supervisor span and this root span.
    def hydrate_span(self, to_wrap, wrapped, instance, args, kwargs, result, span, parent_span = None, ex:Exception = None) -> bool:
        if is_root_agent_name(instance) and "parent.agent.span" in span.attributes:
            agent_request_wrapper = to_wrap.copy()
            agent_request_wrapper["output_processor"] = AGENT_REQUEST
        else:
            agent_request_wrapper = to_wrap
            if hasattr(instance, 'name') and parent_span is not None and not SpanHandler.is_root_span(parent_span):
                parent_span.set_attribute("parent.agent.span", True)
        return super().hydrate_span(agent_request_wrapper, wrapped, instance, args, kwargs, result, span, parent_span, ex)

class LanggraphToolHandler(SpanHandler):
    # LangGraph uses an internal tool to initate delegation to other agents. The method is tool invoke() with tool name as `transfer_to_<agent_name>`.
    # Hence we usea different output processor for tool invoke() to format the span as agentic.delegation.
    def hydrate_span(self, to_wrap, wrapped, instance, args, kwargs, result, span, parent_span = None, ex:Exception = None) -> bool:
        if is_delegation_tool(instance):
            agent_request_wrapper = to_wrap.copy()
            agent_request_wrapper["output_processor"] = AGENT_DELEGATION
        else:
            agent_request_wrapper = to_wrap

        return super().hydrate_span(agent_request_wrapper, wrapped, instance, args, kwargs, result, span, parent_span, ex)
    


from monocle_apptrace.instrumentation.metamodel.gemini import (
    _helper,
)

RETRIEVAL = {
    "type": "retrieval",
    "attributes": [
        [
            {
                "_comment": "Embedding Model",
                "attribute": "name",
                "accessor": lambda arguments: _helper.resolve_from_alias(arguments['kwargs'],
                                                                         ['model'])
            },
            {
                "attribute": "type",
                "accessor": lambda arguments: 'model.embedding.' + _helper.resolve_from_alias(arguments['kwargs'],
                                                                                        ['model'])
            }
        ]
    ],
    "events": [
        {
         "name": "data.input",
         "attributes": [
             {
                 "attribute": "input",
                 "accessor": lambda arguments: _helper.update_input_span_events(arguments['kwargs'])
             }
         ]
         },
        {
            "name": "data.output",
            "attributes": [
                {
                    "attribute": "response",
                    "accessor": lambda arguments: _helper.update_output_span_events(arguments['result'])
                }
            ]
        }

    ]
}



from monocle_apptrace.instrumentation.metamodel.langchain import (
    _helper,
)
from monocle_apptrace.instrumentation.common.utils import get_error_message, resolve_from_alias, get_llm_type, get_status, get_status_code

INFERENCE = {
    "type": "inference.framework",
    "attributes": [
        [
            {
                "_comment": "provider type ,name , deployment , inference_endpoint",
                "attribute": "type",
                "accessor": lambda arguments: 'inference.' + (get_llm_type(arguments['instance']) or 'generic')

            },
            {
                "attribute": "provider_name",
                "accessor": lambda arguments: _helper.extract_provider_name(arguments['instance'])
            },
            {
                "attribute": "deployment",
                "accessor": lambda arguments: resolve_from_alias(arguments['instance'].__dict__, ['engine', 'azure_deployment', 'deployment_name', 'deployment_id', 'deployment'])
            },
            {
                "attribute": "inference_endpoint",
                "accessor": lambda arguments: resolve_from_alias(arguments['instance'].__dict__, ['azure_endpoint', 'api_base', 'endpoint']) or _helper.extract_inference_endpoint(arguments['instance'])
            }
        ],
        [
            {
                "_comment": "LLM Model",
                "attribute": "name",
                "accessor": lambda arguments: resolve_from_alias(arguments['instance'].__dict__, ['model', 'model_name', 'endpoint_name', 'deployment_name', 'model_id'])
            },
            {
                "attribute": "type",
                "accessor": lambda arguments: 'model.llm.' + resolve_from_alias(arguments['instance'].__dict__, ['model', 'model_name', 'endpoint_name', 'deployment_name', 'model_id'])
            }
        ]
    ],
    "events": [
        {"name": "data.input",
         "attributes": [

             {
                 "_comment": "this is instruction and user query to LLM",
                 "attribute": "input",
                 "accessor": lambda arguments: _helper.extract_messages(arguments['args'])
             }
         ]
         },
        {
            "name": "data.output",
            "attributes": [
                {
                    "attribute": "error_code",
                    "accessor": lambda arguments: get_error_message(arguments)
                },
                {
                    "attribute": "response",
                    "accessor": lambda arguments: _helper.extract_assistant_message(arguments)
                }
            ]
        },
        {
            "name": "metadata",
            "attributes": [
                {
                    "_comment": "this is metadata usage from LLM",
                    "accessor": lambda arguments: _helper.update_span_from_llm_response(arguments['result'], arguments['instance'])
                },
                {
                    "attribute": "finish_reason",
                    "accessor": lambda arguments: _helper.extract_finish_reason(arguments)
                },
                {
                    "attribute": "finish_type",
                    "accessor": lambda arguments: _helper.map_finish_reason_to_finish_type(
                        _helper.extract_finish_reason(arguments)
                    )
                }
            ]
        }
    ]
}



from monocle_apptrace.instrumentation.common.wrapper import atask_wrapper, task_wrapper
from monocle_apptrace.instrumentation.metamodel.lambdafunc.entities.http import LAMBDA_HTTP_PROCESSOR

LAMBDA_HTTP_METHODS = [
    {
        "package": "monocle_apptrace.instrumentation.metamodel.lambdafunc.wrapper",
        "object": "LambdaFunctionRouteWrapper",
        "method": "run_async",
        "span_name": "lambda_function_route",
        "wrapper_method": atask_wrapper,
        "span_handler": "lambda_func_handler",
        "output_processor": LAMBDA_HTTP_PROCESSOR
    },
    {
        "package": "monocle_apptrace.instrumentation.metamodel.lambdafunc.wrapper",
        "object": "LambdaFunctionRouteWrapper",
        "method": "run_sync",
        "span_name": "lambda_function_route",
        "wrapper_method": task_wrapper,
        "span_handler": "lambda_func_handler",
        "output_processor": LAMBDA_HTTP_PROCESSOR
    }
]



from monocle_apptrace.instrumentation.common.wrapper import atask_wrapper, task_wrapper
from monocle_apptrace.instrumentation.metamodel.llamaindex.entities.inference import (
    INFERENCE,
)
from monocle_apptrace.instrumentation.metamodel.llamaindex.entities.agent import AGENT, TOOLS, AGENT_REQUEST
from monocle_apptrace.instrumentation.metamodel.llamaindex.entities.retrieval import (
    RETRIEVAL,
)


LLAMAINDEX_METHODS = [
    {
        "package": "llama_index.core.indices.base_retriever",
        "object": "BaseRetriever",
        "method": "retrieve",
        "wrapper_method": task_wrapper,
        "output_processor": RETRIEVAL
    },
    {
        "package": "llama_index.core.indices.base_retriever",
        "object": "BaseRetriever",
        "method": "aretrieve",
        "wrapper_method": atask_wrapper,
        "output_processor": RETRIEVAL
    },
    {
        "package": "llama_index.core.base.base_query_engine",
        "object": "BaseQueryEngine",
        "method": "query",
        "wrapper_method": task_wrapper
    },
    {
        "package": "llama_index.core.base.base_query_engine",
        "object": "BaseQueryEngine",
        "method": "aquery",
        "wrapper_method": atask_wrapper
    },
    {
        "package": "llama_index.core.llms.custom",
        "object": "CustomLLM",
        "method": "chat",
        "wrapper_method": task_wrapper,
        "output_processor": INFERENCE
    },
    {
        "package": "llama_index.core.llms.custom",
        "object": "CustomLLM",
        "method": "achat",
        "wrapper_method": atask_wrapper,
        "output_processor": INFERENCE,
        
    },
    {
        "package": "llama_index.llms.openai.base",
        "object": "OpenAI",
        "method": "chat",
        "wrapper_method": task_wrapper,
        "output_processor": INFERENCE
    },
    {
        "package": "llama_index.llms.openai.base",
        "object": "OpenAI",
        "method": "achat",
        "wrapper_method": atask_wrapper,
        "output_processor": INFERENCE
    },
    {
        "package": "llama_index.llms.mistralai.base",
        "object": "MistralAI",
        "method": "chat",
        "wrapper_method": task_wrapper,
        "output_processor": INFERENCE
    },
    {
        "package": "llama_index.llms.mistralai.base",
        "object": "MistralAI",
        "method": "achat",
        "wrapper_method": atask_wrapper,
        "output_processor": INFERENCE
    },
    {
        "package": "llama_index.core.agent.workflow.multi_agent_workflow",
        "object": "AgentWorkflow",
        "method": "run",
        "span_handler": "llamaindex_agent_handler",
        "wrapper_method": atask_wrapper,
        "output_processor": AGENT_REQUEST
    },
    {
        "package": "llama_index.core.agent",
        "object": "ReActAgent",
        "method": "chat",
        "wrapper_method": task_wrapper,
        "output_processor": AGENT
    },
    {
        "package": "llama_index.core.agent",
        "object": "ReActAgent",
        "method": "achat",
        "wrapper_method": atask_wrapper,
        "output_processor": AGENT
    },
    {
        "package": "llama_index.core.agent.workflow.function_agent",
        "object": "FunctionAgent",
        "method": "finalize",
        "wrapper_method": atask_wrapper,
        "output_processor": AGENT
    },
    {
        "package": "llama_index.core.agent.workflow.function_agent",
        "object": "FunctionAgent",
        "method": "take_step",
        "span_handler": "llamaindex_agent_handler",
        "wrapper_method": atask_wrapper
    },
    {
        "package": "llama_index.core.tools.function_tool",
        "object": "FunctionTool",
        "method": "call",
        "span_handler": "llamaindex_single_agent_tool_handler",
        "wrapper_method": task_wrapper,
        "output_processor": TOOLS
    },
    {
        "package": "llama_index.core.tools.function_tool",
        "object": "FunctionTool",
        "method": "acall",
        "span_handler": "llamaindex_single_agent_tool_handler",
        "wrapper_method": atask_wrapper,
        "output_processor": TOOLS
    },
    {
        "package": "llama_index.core.agent.workflow.multi_agent_workflow",
        "object": "AgentWorkflow",
        "method": "_call_tool",
        "span_handler": "llamaindex_tool_handler",
        "wrapper_method": atask_wrapper,
        "output_processor": TOOLS
    },
    {
        "package": "llama_index.llms.anthropic",
        "object": "Anthropic",
        "method": "chat",
        "wrapper_method": task_wrapper,
        "output_processor": INFERENCE
    },
    {
        "package": "llama_index.llms.anthropic",
        "object": "Anthropic",
        "method": "achat",
        "wrapper_method": atask_wrapper,
        "output_processor": INFERENCE
    },
    {
        "package": "llama_index.llms.gemini",
        "object": "Gemini",
        "method": "chat",
        "wrapper_method": task_wrapper,
        "output_processor": INFERENCE
    },
    {
        "package": "llama_index.llms.gemini",
        "object": "Gemini",
        "method": "achat",
        "wrapper_method": atask_wrapper,
        "output_processor": INFERENCE
    }
]



from monocle_apptrace.instrumentation.metamodel.openai import (
    _helper,
)
from monocle_apptrace.instrumentation.common.utils import resolve_from_alias

RETRIEVAL = {
    "type": "embedding",
    "attributes": [
        [
            {
                "_comment": "LLM Model",
                "attribute": "name",
                "accessor": lambda arguments: resolve_from_alias(arguments['kwargs'], ['model', 'model_name', 'endpoint_name', 'deployment_name'])
            },
            {
                "attribute": "type",
                "accessor": lambda arguments: 'model.embedding.' + resolve_from_alias(arguments['kwargs'], ['model', 'model_name', 'endpoint_name', 'deployment_name'])
            }
        ]
    ],
    "events": [
        {
             "name": "data.input",
             "attributes": [
                 {
                     "_comment": "this is instruction and user query to LLM",
                     "attribute": "input",
                     "accessor": lambda arguments: _helper.update_input_span_events(arguments['kwargs'])
                 }
         ]
         },
        {
            "name": "data.output",
            "attributes": [
                {
                    "_comment": "this is result from LLM",
                    "attribute": "response",
                    "accessor": lambda arguments: _helper.update_output_span_events(arguments['result'])
                }
            ]
        }
    ]
}



import logging
from threading import local
from monocle_apptrace.instrumentation.common.utils import extract_http_headers, clear_http_scopes, try_option, Option, MonocleSpanException
from monocle_apptrace.instrumentation.common.span_handler import SpanHandler
from monocle_apptrace.instrumentation.common.constants import HTTP_SUCCESS_CODES
from urllib.parse import unquote, urlparse, ParseResult


logger = logging.getLogger(__name__)
MAX_DATA_LENGTH = 1000

def get_url(kwargs) -> ParseResult:
    url_str = try_option(getattr, kwargs['req'], 'url')
    url = url_str.unwrap_or(None)
    if url is not None:
        return urlparse(url)
    else:
        return None

def get_function_name(kwargs) -> str:
    context = kwargs.get('context', None)
    if context is not None and hasattr(context, 'function_name'):
        return context.function_name
    return ""
    

def get_route(kwargs) -> str:
    url:ParseResult = get_url(kwargs)
    if url is not None:
        return url.path

def get_method(kwargs) -> str:
#    return args[0]['method'] if 'method' in args[0] else ""
    http_method: Option[str] = try_option(getattr, kwargs['req'], 'method')
    return http_method.unwrap_or("")

def get_params(kwargs) -> dict:
    url:ParseResult = get_url(kwargs)
    if url is not None:
        return unquote(url.query)

def get_body(kwargs) -> dict:
    if hasattr(kwargs['req'], 'get_body'):
        response = kwargs.get_body()
        if isinstance(response, bytes):
            response = response.decode('utf-8', errors='ignore')
    else:
        response = ""
    return response

def extract_response(result) -> str:
    if hasattr(result, 'get_body'):
        response = result.get_body() #  text[0:max(result.text.__len__(), MAX_DATA_LENGTH)]
        if isinstance(response, bytes):
            response = response.decode('utf-8', errors='ignore')
    else:
        response = ""
    return response

def extract_status(result) -> str:
    status = f"{result.status_code}" if hasattr(result, 'status_code') else ""
    if status not in HTTP_SUCCESS_CODES:
        error_message = extract_response(result)
        raise MonocleSpanException(f"error: {status} - {error_message}")
    return status

def azure_func_pre_tracing(kwargs):
    headers = kwargs['req'].headers if hasattr(kwargs['req'], 'headers') else {}
    return extract_http_headers(headers)

def azure_func_post_tracing(token):
    clear_http_scopes(token)

class azureSpanHandler(SpanHandler):

    def pre_tracing(self, to_wrap, wrapped, instance, args, kwargs):
        return azure_func_pre_tracing(kwargs)
    
    def post_tracing(self, to_wrap, wrapped, instance, args, kwargs, return_value, token):
        azure_func_post_tracing(token)



import json
import logging
from typing import Any, Dict, Optional
from urllib.parse import urlparse
from monocle_apptrace.instrumentation.common.utils import (
    get_json_dumps,
    get_exception_message,
    get_status_code,
)
from monocle_apptrace.instrumentation.metamodel.finish_types import map_azure_ai_inference_finish_reason_to_finish_type

logger = logging.getLogger(__name__)


def extract_messages(args_or_kwargs: Any) -> str:
    """Extract messages from azure-ai-inference request arguments."""
    try:
        messages = []
        if "instructions" in args_or_kwargs:
            messages.append({"instructions": args_or_kwargs.get("instructions", {})})
        if "input" in args_or_kwargs:
            messages.append({"input": args_or_kwargs.get("input", {})})
        if "messages" in args_or_kwargs and len(args_or_kwargs["messages"]) > 0:
            for msg in args_or_kwargs["messages"]:
                if msg.get("content") and msg.get("role"):
                    messages.append({msg["role"]: msg["content"]})

        return [get_json_dumps(message) for message in messages]
    except Exception as e:
        logger.warning("Warning: Error occurred in extract_messages: %s", str(e))
        return []


def extract_inference_endpoint(instance: Any) -> str:
    """Extract inference endpoint from azure-ai-inference client instance."""
    try:
        return instance._config.endpoint
    except Exception as e:
        logger.warning(
            "Warning: Error occurred in extract_inference_endpoint: %s", str(e)
        )
        return ""


def extract_embeddings_input(args_or_kwargs: Any) -> str:
    """Extract input text from azure-ai-inference embeddings request."""
    try:
        # Handle both args and kwargs scenarios
        if isinstance(args_or_kwargs, dict):
            if "input" in args_or_kwargs:
                input_data = args_or_kwargs["input"]
            elif len(args_or_kwargs) > 0:
                first_arg = list(args_or_kwargs.values())[0]
                if isinstance(first_arg, dict) and "input" in first_arg:
                    input_data = first_arg["input"]
                else:
                    input_data = first_arg
            else:
                return ""
        elif hasattr(args_or_kwargs, "__iter__") and len(args_or_kwargs) > 0:
            first_arg = args_or_kwargs[0]
            if hasattr(first_arg, "get") and "input" in first_arg:
                input_data = first_arg["input"]
            else:
                input_data = first_arg
        else:
            return ""

        # Format input for display
        if isinstance(input_data, (list, tuple)):
            return " | ".join(str(item) for item in input_data)
        else:
            return str(input_data)
    except Exception as e:
        logger.warning(
            "Warning: Error occurred in extract_embeddings_input: %s", str(e)
        )
        return ""


def extract_assistant_message(arguments: Dict[str, Any]) -> str:
    """Extract assistant response from azure-ai-inference completion result."""
    try:
        # Check for exception first
        if arguments.get("exception") is not None:
            return get_exception_message(arguments)

        result = arguments.get("result")
        role = "assistant"
        messages = []
        if not result:
            return ""
        if hasattr(result, "output_text"):
            # If the result has output_text attribute
            role = getattr(result, "role", role)
            if "assistant" in role.lower():
                # If the role is assistant, we can assume it's a chat completion
                role = "assistant"
            messages.append({role: result.output_text})
        if (hasattr(result, "choices")
            and result.choices
            and result.choices[0].message
            and result.choices[0].message.content
        ):
            role = getattr(result.choices[0].message, "role", role)
            if "assistant" in role.lower():
                # If the role is assistant, we can assume it's a chat completion
                role = "assistant"
            # If the result is a chat completion with content
            messages.append({role: result.choices[0].message.content})
        return get_json_dumps(messages[0]) if messages else ""
    except Exception as e:
        logger.warning(
            "Warning: Error occurred in extract_assistant_message: %s", str(e)
        )
        return ""


def extract_embeddings_output(arguments: Dict[str, Any]) -> str:
    """Extract embeddings from azure-ai-inference embeddings result."""
    try:
        result = arguments.get("result")
        if not result:
            return ""

        if hasattr(result, "data") and result.data:
            # Format as summary of embeddings data
            embeddings_info = []
            for i, item in enumerate(result.data):
                if hasattr(item, "embedding") and hasattr(item, "index"):
                    embedding_length = len(item.embedding) if item.embedding else 0
                    embeddings_info.append(
                        f"index={item.index}, embedding=[{embedding_length} dimensions]"
                    )
            return " | ".join(embeddings_info)

        return str(result)
    except Exception as e:
        logger.warning(
            "Warning: Error occurred in extract_embeddings_output: %s", str(e)
        )
        return ""


def update_span_from_llm_response(result: Any, instance: Any = None) -> Dict[str, Any]:
    """Extract usage metadata from azure-ai-inference response."""
    try:
        attributes = {}

        # Handle streaming responses with accumulated usage data
        if hasattr(result, "usage") and result.usage:
            usage = result.usage
            if hasattr(usage, "completion_tokens"):
                attributes["completion_tokens"] = usage.completion_tokens
            if hasattr(usage, "prompt_tokens"):
                attributes["prompt_tokens"] = usage.prompt_tokens
            if hasattr(usage, "total_tokens"):
                attributes["total_tokens"] = usage.total_tokens

        # Handle regular response usage
        elif hasattr(result, "usage"):
            usage = result.usage
            if hasattr(usage, "completion_tokens"):
                attributes["completion_tokens"] = usage.completion_tokens
            if hasattr(usage, "prompt_tokens"):
                attributes["prompt_tokens"] = usage.prompt_tokens
            if hasattr(usage, "total_tokens"):
                attributes["total_tokens"] = usage.total_tokens

        # Extract model information if available
        if hasattr(result, "model"):
            attributes["model"] = result.model

        return attributes
    except Exception as e:
        logger.warning(
            "Warning: Error occurred in update_span_from_llm_response: %s", str(e)
        )
        return {}


def get_model_name(arguments: Dict[str, Any]) -> str:
    """Extract model name from azure-ai-inference request arguments."""
    try:

        # Try to get from instance
        instance = arguments.get("instance")
        if arguments.get('kwargs') and arguments.get('kwargs').get('model'):
            return arguments['kwargs'].get('model')
        if instance and hasattr(instance, "_config") and hasattr(instance._config, "model"):
            return instance._config.endpoint.split("/")[-1]

        return ""
    except Exception as e:
        logger.warning("Warning: Error occurred in get_model_name: %s", str(e))
        return ""


def get_inference_type(arguments) -> str:
    instance = arguments.get("instance")
    if instance and hasattr(instance, "_config") and hasattr(instance._config, "endpoint"):
        endpoint = instance._config.endpoint
        try:
            parsed = urlparse(endpoint)
            hostname = parsed.hostname or endpoint
            hostname = hostname.lower()
        except Exception:
            hostname = str(endpoint).lower()
        if hostname.endswith("services.ai.azure.com"):
            return "azure_ai_inference"
        if hostname.endswith("openai.azure.com"):
            return "azure_openai"
    return "azure_ai_inference"


def get_provider_name(instance: Any) -> str:
    """Extract provider name from azure-ai-inference client instance."""
    try:
        # extract hostname from instance._config.endpoint
        # https://okahu-openai-dev.openai.azure.com/openai/deployments/kshitiz-gpt => okahu-openai-dev.openai.azure.com
        endpoint = instance._config.endpoint
        if endpoint:
            # Extract the hostname part
            provider_name = endpoint.split("/")[2] if "/" in endpoint else endpoint
            return provider_name
    except Exception as e:
        logger.warning("Warning: Error occurred in get_provider_name: %s", str(e))
        return "azure_ai_inference"


def extract_finish_reason(arguments: Dict[str, Any]) -> Optional[str]:
    """Extract finish_reason from Azure AI Inference response."""
    try:
        # Handle exception cases first
        if arguments.get("exception") is not None:
            ex = arguments["exception"]
            if hasattr(ex, "message") and isinstance(ex.message, str):
                message = ex.message
                if "content_filter" in message.lower():
                    return "content_filter"
            return "error"
        
        result = arguments.get("result")
        if result is None:
            return None
            
        # Check various possible locations for finish_reason in Azure AI Inference responses
        
        # Direct finish_reason attribute
        if hasattr(result, "finish_reason") and result.finish_reason:
            return result.finish_reason
            
        # Check for choices structure (OpenAI-compatible format)
        if hasattr(result, "choices") and result.choices:
            choice = result.choices[0]
            if hasattr(choice, "finish_reason") and choice.finish_reason:
                return choice.finish_reason
        
        # Check for additional metadata or response attributes
        if hasattr(result, "additional_kwargs") and result.additional_kwargs:
            kwargs = result.additional_kwargs
            if isinstance(kwargs, dict):
                for key in ["finish_reason", "stop_reason"]:
                    if key in kwargs:
                        return kwargs[key]
        
        # Check for response metadata
        if hasattr(result, "response_metadata") and result.response_metadata:
            metadata = result.response_metadata
            if isinstance(metadata, dict):
                for key in ["finish_reason", "stop_reason"]:
                    if key in metadata:
                        return metadata[key]
        
        # Check for streaming response with accumulated finish reason
        if hasattr(result, "type") and result.type == "stream":
            # For streaming responses, default to stop if completed successfully
            return "stop"
        
        # If no specific finish reason found, infer from status
        status_code = get_status_code(arguments)
        if status_code == 'success':
            return "stop"  # Default success finish reason
        elif status_code == 'error':
            return "error"
            
    except Exception as e:
        logger.warning("Warning: Error occurred in extract_finish_reason: %s", str(e))
        return None
    
    return None


def map_finish_reason_to_finish_type(finish_reason: Optional[str]) -> Optional[str]:
    """Map Azure AI Inference finish_reason to finish_type."""
    return map_azure_ai_inference_finish_reason_to_finish_type(finish_reason)



from typing import Dict, Any, List
import os
import logging, warnings
from importlib import import_module
from opentelemetry.sdk.trace.export import SpanExporter, ConsoleSpanExporter
from monocle_apptrace.exporters.exporter_processor import LambdaExportTaskProcessor, is_aws_lambda_environment
from monocle_apptrace.exporters.file_exporter import FileSpanExporter

logger = logging.getLogger(__name__)

monocle_exporters: Dict[str, Any] = {
    "s3": {"module": "monocle_apptrace.exporters.aws.s3_exporter", "class": "S3SpanExporter"},
    "blob": {"module": "monocle_apptrace.exporters.azure.blob_exporter", "class": "AzureBlobSpanExporter"},
    "okahu": {"module": "monocle_apptrace.exporters.okahu.okahu_exporter", "class": "OkahuSpanExporter"},
    "file": {"module": "monocle_apptrace.exporters.file_exporter", "class": "FileSpanExporter"},
    "memory": {"module": "opentelemetry.sdk.trace.export.in_memory_span_exporter", "class": "InMemorySpanExporter"},
    "console": {"module": "opentelemetry.sdk.trace.export", "class": "ConsoleSpanExporter"}
}


def get_monocle_exporter(exporters_list:str=None) -> List[SpanExporter]:
    # Retrieve the MONOCLE_EXPORTER environment variable and split it into a list
    if exporters_list:
        exporter_names = exporters_list.split(",")
    else:
        exporter_names = os.environ.get("MONOCLE_EXPORTER", "file").split(",")
    exporters = []
    
    # Create task processor for AWS Lambda environment
    task_processor = LambdaExportTaskProcessor() if is_aws_lambda_environment() else None

    for exporter_name in exporter_names:
        exporter_name = exporter_name.strip()
        try:
            exporter_class_path = monocle_exporters[exporter_name]
        except KeyError:
            warnings.warn(f"Unsupported Monocle span exporter '{exporter_name}', skipping.")
            continue
        try:
            exporter_module = import_module(exporter_class_path["module"])
            exporter_class = getattr(exporter_module, exporter_class_path["class"])
            # Pass task_processor to all exporters when in AWS Lambda environment
            if task_processor is not None and exporter_module.__name__.startswith("monocle_apptrace"):
                exporters.append(exporter_class(task_processor=task_processor))
            else:
                exporters.append(exporter_class())
        except Exception as ex:
            warnings.warn(
                f"Unable to initialize Monocle span exporter '{exporter_name}', error: {ex}. Using ConsoleSpanExporter as a fallback.")
            exporters.append(ConsoleSpanExporter())
            continue

    # If no exporters were created, default to FileSpanExporter
    if not exporters:
        logger.debug("No valid Monocle span exporters configured. Defaulting to FileSpanExporter.")
        exporters.append(FileSpanExporter())

    return exporters



from monocle_apptrace.instrumentation.common.wrapper import task_wrapper
from monocle_apptrace.instrumentation.metamodel.requests.entities.http import REQUEST_HTTP_PROCESSOR

REQUESTS_METHODS = [
    {
        "package": "requests.sessions",
        "object": "Session",
        "method": "request",
        "wrapper_method": task_wrapper,
        "span_handler":"request_handler",
        "output_processor": REQUEST_HTTP_PROCESSOR
    }
]


"""
This module provides utility functions for extracting system, user,
and assistant messages from various input formats.
"""

import logging
import json
from io import BytesIO
from functools import wraps
from monocle_apptrace.instrumentation.common.span_handler import SpanHandler
from monocle_apptrace.instrumentation.common.utils import ( get_exception_message, get_json_dumps, get_status_code,)
from monocle_apptrace.instrumentation.metamodel.finish_types import map_bedrock_finish_reason_to_finish_type
logger = logging.getLogger(__name__)


def extract_messages(args):
    """Extract system and user messages"""
    try:
        messages = []
        if args and isinstance(args, dict) and len(args) > 0:
            if 'Body' in args and isinstance(args['Body'], str):
                data = json.loads(args['Body'])
                question = data.get("question")
                messages.append(question)
            if 'messages' in args and isinstance(args['messages'], list):
                role = args['messages'][0]['role']
                user_message = extract_query_from_content(args['messages'][0]['content'][0]['text'])
                messages.append({role: user_message})
        return [get_json_dumps(message) for message in messages]
    except Exception as e:
        logger.warning("Warning: Error occurred in extract_messages: %s", str(e))
        return []

def get_exception_status_code(arguments):
    if arguments['exception'] is not None and hasattr(arguments['exception'], 'response') and arguments['exception'].response is not None:
        if "ResponseMetadata" in arguments['exception'].response and "HTTPStatusCode" in arguments['exception'].response["ResponseMetadata"]:
            return arguments['exception'].response["ResponseMetadata"]["HTTPStatusCode"]
    elif arguments['exception'] is not None:
        return 'error'
    else:
        return 'success'

def extract_assistant_message(arguments):
    try:
        status = get_status_code(arguments)
        messages = []
        role = "assistant"
        if status == 'success':
            if "Body" in arguments['result'] and hasattr(arguments['result']['Body'], "_raw_stream"):
                raw_stream = getattr(arguments['result']['Body'], "_raw_stream")
                if hasattr(raw_stream, "data"):
                    response_bytes = getattr(raw_stream, "data")
                    response_str = response_bytes.decode('utf-8')
                    response_dict = json.loads(response_str)
                    arguments['result']['Body'] = BytesIO(response_bytes)
                    messages.append({role: response_dict["answer"]})
            if "output" in arguments['result']:
                output = arguments['result'].get("output", {})
                message = output.get("message", {})
                content = message.get("content", [])
                if isinstance(content, list) and len(content) > 0 and "text" in content[0]:
                    reply = content[0]["text"]
                    messages.append({role: reply})
        else:
            if arguments["exception"] is not None:
                return get_exception_message(arguments)
            elif hasattr(arguments["result"], "error"):
                return arguments["result"].error
        return get_json_dumps(messages[0]) if messages else ""
    except Exception as e:
        logger.warning("Warning: Error occurred in extract_assistant_message: %s", str(e))
        return []


def extract_query_from_content(content):
    try:
        query_prefix = "Query:"
        answer_prefix = "Answer:"
        query_start = content.find(query_prefix)
        if query_start == -1:
            return None

        query_start += len(query_prefix)
        answer_start = content.find(answer_prefix, query_start)
        if answer_start == -1:
            query = content[query_start:].strip()
        else:
            query = content[query_start:answer_start].strip()
        return query
    except Exception as e:
        logger.warning("Warning: Error occurred in extract_query_from_content: %s", str(e))
        return ""


def resolve_from_alias(my_map, alias):
    """Find a alias that is not none from list of aliases"""

    for i in alias:
        if i in my_map.keys():
            return my_map[i]
    return None

def update_span_from_llm_response(response, instance):
    meta_dict = {}
    if response is not None and isinstance(response, dict) and "usage" in response:
        token_usage = response["usage"]

        if token_usage is not None:
            temperature = instance.__dict__.get("temperature", None)
            meta_dict.update({"temperature": temperature})
            meta_dict.update({"completion_tokens": resolve_from_alias(token_usage,["completion_tokens","output_tokens","outputTokens"])})
            meta_dict.update({"prompt_tokens": resolve_from_alias(token_usage,["prompt_tokens","input_tokens","inputTokens"])})
            meta_dict.update({"total_tokens": resolve_from_alias(token_usage,["total_tokens","totalTokens"])})
    return meta_dict


def extract_finish_reason(arguments):
    """Extract finish_reason/stopReason from Bedrock response."""
    try:
        # Handle exception cases first
        if arguments.get("exception") is not None:
            return "error"
        
        result = arguments.get("result")
        if result is None:
            return None
            
        # Check various possible locations for finish_reason in Bedrock responses
        
        # Direct stopReason attribute (Bedrock Converse API)
        if "stopReason" in result:
            return result["stopReason"]
            
        # Check for completionReason (some Bedrock models)
        if "completionReason" in result:
            return result["completionReason"]
            
        # Check for output structure (Bedrock Converse API)
        if "output" in result and "message" in result["output"]:
            message = result["output"]["message"]
            if "stopReason" in message:
                return message["stopReason"]
                
        # Check for nested result structure
        if "result" in result:
            nested_result = result["result"]
            if "stopReason" in nested_result:
                return nested_result["stopReason"]
            if "completionReason" in nested_result:
                return nested_result["completionReason"]
                
        # Check for streaming response accumulated finish reason
        if "type" in result and result["type"] == "stream":
            if "stopReason" in result:
                return result["stopReason"]
                
        # Check for response metadata
        if "ResponseMetadata" in result:
            metadata = result["ResponseMetadata"]
            if "stopReason" in metadata:
                return metadata["stopReason"]
                
        # Check for Body content (for some Bedrock responses)
        if "Body" in result:
            body = result["Body"]
            if hasattr(body, "_raw_stream"):
                raw_stream = getattr(body, "_raw_stream")
                if hasattr(raw_stream, "data"):
                    response_bytes = getattr(raw_stream, "data")
                    response_str = response_bytes.decode('utf-8')
                    try:
                        response_dict = json.loads(response_str)
                        if "stopReason" in response_dict:
                            return response_dict["stopReason"]
                        if "completionReason" in response_dict:
                            return response_dict["completionReason"]
                    except json.JSONDecodeError:
                        pass
                        
        # If no specific finish reason found, infer from status
        status_code = get_status_code(arguments)
        if status_code == 'success':
            return "end_turn"  # Default successful completion
        elif status_code == 'error':
            return "error"
            
    except Exception as e:
        logger.warning("Warning: Error occurred in extract_finish_reason: %s", str(e))
        return None
    
    return None


def map_finish_reason_to_finish_type(finish_reason):
    """Map Bedrock finish_reason/stopReason to finish_type."""
    return map_bedrock_finish_reason_to_finish_type(finish_reason)



from monocle_apptrace.instrumentation.common.utils import with_tracer_wrapper
from opentelemetry.context import attach, set_value, get_value, detach
from monocle_apptrace.instrumentation.common.utils import resolve_from_alias
import logging
import json

logger = logging.getLogger(__name__)


def log(arguments):
    print(f"Arguments: {arguments}")


def get_output_text(arguments):
    # arguments["result"].content[0].text
    if (
        "result" in arguments
        and hasattr(arguments["result"], "tools")
        and isinstance(arguments["result"].tools, list)
    ):
        tools = []
        for tool in arguments["result"].tools:
            if hasattr(tool, "name"):
                tools.append(tool.name)
        return tools
    if (
        "result" in arguments
        and hasattr(arguments["result"], "content")
        and isinstance(arguments["result"].content, list)
    ):
        ret_val = []
        for content in arguments["result"].content:
            if hasattr(content, "text"):
                ret_val.append(content.text)
        return ret_val


def get_name(arguments):
    """Get the name of the tool from the instance."""

    args = arguments["args"]
    if (
        args
        and hasattr(args[0], "root")
        and hasattr(args[0].root, "params")
        and hasattr(args[0].root.params, "name")
    ):
        # If the first argument has a root with params and name, return that name
        return args[0].root.params.name


def get_type(arguments):
    """Get the type of the tool from the instance."""
    args = arguments["args"]
    if args and hasattr(args[0], "root") and hasattr(args[0].root, "method"):
        # If the first argument has a root with a method, return that method's name
        return args[0].root.method


def get_params_arguments(arguments):
    """Get the params of the tool from the instance."""

    args = arguments["args"]
    if (
        args
        and hasattr(args[0], "root")
        and hasattr(args[0].root, "params")
        and hasattr(args[0].root.params, "arguments")
    ):
        # If the first argument has a root with params and arguments, return those arguments
        try:
            return json.dumps(args[0].root.params.arguments)
        except (TypeError, ValueError) as e:
            logger.error(f"Error serializing arguments: {e}")
            return str(args[0].root.params.arguments)


def get_url(arguments):
    """Get the URL of the tool from the instance."""
    url = get_value("mcp.url", None)

    return url

# this extracts the url from the langchain mcp adapter tools and attaches it to the context.
@with_tracer_wrapper
def langchain_mcp_wrapper(
    tracer: any, handler: any, to_wrap, wrapped, instance, source_path, args, kwargs
):
    return_value = None
    try:
        return_value = wrapped(*args, **kwargs)
        return return_value
    finally:
        if (
            return_value
            and hasattr(return_value, "coroutine")
            and kwargs.get("connection", None)
        ):
            try:
                # extract the URL from the connection and attach it to the context
                url = kwargs.get("connection").get("url", None)
                if url:
                    # wrap coroutine methods and attach the URL to the context

                    original_coroutine = return_value.coroutine

                    async def wrapped_coroutine(*args1, **kwargs1):
                        token = None
                        try:
                            token = attach(set_value("mcp.url", url))
                            return await original_coroutine(*args1, **kwargs1)
                        finally:
                            detach(token)

                    return_value.coroutine = wrapped_coroutine

            except Exception as e:
                pass



from functools import wraps
import inspect

def monocle_trace_lambda_function_route(func):
    if inspect.iscoroutinefunction(func):
        @wraps(func)
        async def wrapper(*args, **kwargs):
            return await LambdaFunctionRouteWrapper.run_async(func, *args, **kwargs)
        return wrapper
    else:
        @wraps(func)
        def wrapper(*args, **kwargs):
            return LambdaFunctionRouteWrapper.run_sync(func, *args, **kwargs)
        return wrapper

class LambdaFunctionRouteWrapper:
    @staticmethod
    async def run_async(func, *args, **kwargs):
        return await func(*args, **kwargs)

    @staticmethod
    def run_sync(func, *args, **kwargs):
        return func(*args, **kwargs)


import logging
from monocle_apptrace.instrumentation.common.utils import (
    get_exception_message,
    get_json_dumps,
    get_status_code,
)
from monocle_apptrace.instrumentation.metamodel.finish_types import (
    map_gemini_finish_reason_to_finish_type,
    GEMINI_FINISH_REASON_MAPPING
)

logger = logging.getLogger(__name__)

def resolve_from_alias(my_map, alias):
    """Find a alias that is not none from list of aliases"""

    for i in alias:
        if i in my_map.keys():
            return my_map[i]
    return None

def extract_messages(kwargs):
    """Extract system and user messages"""
    try:
        messages = []
        config = kwargs.get('config')
        if config and hasattr(config, 'system_instruction'):
            system_instructions = getattr(config, 'system_instruction', None)
            if system_instructions:
                messages.append({'system': system_instructions})

        contents = kwargs.get('contents')
        if isinstance(contents, list):
            for content in contents:
                if hasattr(content, 'parts') and getattr(content, 'parts'):
                    part = content.parts[0]
                    if hasattr(part, 'text'):
                        messages.append({getattr(content, 'role', 'user'): part.text})
        elif isinstance(contents, str):
            messages.append({'user': contents})

        return [get_json_dumps(message) for message in messages]
    except Exception as e:
        logger.warning("Warning: Error occurred in extract_messages: %s", str(e))
        return []

def extract_assistant_message(arguments):
    try:
        status = get_status_code(arguments)
        messages = []
        role = "assistant"
        if hasattr(arguments['result'], "candidates") and len(arguments['result'].candidates) > 0 and hasattr(arguments['result'].candidates[0], "content") and hasattr(arguments['result'].candidates[0].content, "role"):
                role = arguments["result"].candidates[0].content.role
        if status == 'success':
            if hasattr(arguments['result'], "text") and len(arguments['result'].text):
                messages.append({role: arguments['result'].text})
        else:
            if arguments["exception"] is not None:
                return get_exception_message(arguments)
            elif hasattr(arguments["result"], "error"):
                return arguments["result"].error
        return get_json_dumps(messages[0]) if messages else ""
    except (IndexError, AttributeError) as e:
        logger.warning("Warning: Error occurred in extract_assistant_message: %s", str(e))
        return None

def update_input_span_events(kwargs):
    if 'contents' in kwargs and isinstance(kwargs['contents'], list) and len(kwargs['contents']) > 0:
        query = kwargs['contents'][0]
        return query

def update_output_span_events(results):
    if hasattr(results,'embeddings') and isinstance(results.embeddings, list) and len(results.embeddings) > 0:
        embeddings = results.embeddings[0]
        if hasattr(embeddings, 'values') and isinstance(embeddings.values, list) and len(embeddings.values) > 100:
            output = str(results.embeddings[0].values[:100]) + "..."
            return output

def extract_inference_endpoint(instance):
    try:
        if hasattr(instance,'_api_client') and hasattr(instance._api_client, '_http_options'):
            if hasattr(instance._api_client._http_options,'base_url'):
                return instance._api_client._http_options.base_url
    except Exception as e:
        logger.warning("Warning: Error occurred in inference endpoint: %s", str(e))
        return []

def update_span_from_llm_response(response, instance):
    meta_dict = {}
    if response is not None and hasattr(response, "usage_metadata") and response.usage_metadata is not None:
        token_usage = response.usage_metadata
        if token_usage is not None:
            meta_dict.update({"completion_tokens": token_usage.candidates_token_count})
            meta_dict.update({"prompt_tokens": token_usage.prompt_token_count })
            meta_dict.update({"total_tokens": token_usage.total_token_count})
    return meta_dict

def extract_finish_reason(arguments):
    """Extract finish_reason from Gemini response"""
    try:
        if arguments["exception"] is not None:
            return None
            
        response = arguments["result"]
        
        # Handle Gemini response structure
        if (response is not None and 
            hasattr(response, "candidates") and 
            len(response.candidates) > 0 and 
            hasattr(response.candidates[0], "finish_reason")):
            return response.candidates[0].finish_reason
            
    except (IndexError, AttributeError) as e:
        logger.warning("Warning: Error occurred in extract_finish_reason: %s", str(e))
        return None
    return None

def map_finish_reason_to_finish_type(finish_reason):
    """Map Gemini finish_reason to finish_type based on the possible errors mapping"""
    return map_gemini_finish_reason_to_finish_type(finish_reason)



from os import environ
from monocle_apptrace.instrumentation.common.constants import TRACE_PROPOGATION_URLS
allowed_url_str = environ.get(TRACE_PROPOGATION_URLS, "")
allowed_urls:list[str] = [] if allowed_url_str == "" else allowed_url_str.split(',')